"""
Generate samples from a regular grid.

Example usage
-------------

Basic usage::

    >>> print(numpy.around(create_grid_samples(order=2), 4))
    [[0.3333 0.6667]]
    >>> print(numpy.around(create_grid_samples(order=5), 4))
    [[0.1667 0.3333 0.5    0.6667 0.8333]]

Certain orders are nested::

    >>> print(numpy.around(create_grid_samples(order=3), 4))
    [[0.25 0.5  0.75]]
    >>> print(numpy.around(create_grid_samples(order=7), 4))
    [[0.125 0.25  0.375 0.5   0.625 0.75  0.875]]

Create nested samples directly with the dedicated function::

    >>> print(numpy.around(create_nested_grid_samples(order=2), 4))
    [[0.25 0.5  0.75]]
    >>> print(numpy.around(create_nested_grid_samples(order=3), 4))
    [[0.125 0.25  0.375 0.5   0.625 0.75  0.875]]

Multivariate usage::

    >>> print(numpy.around(create_grid_samples(order=2, dim=2), 4))
    [[0.3333 0.3333 0.6667 0.6667]
     [0.3333 0.6667 0.3333 0.6667]]
"""
import numpy

import chaospy.quad


def create_grid_samples(order, dim=1):
    """
    Create samples from a regular grid.

    Args:
        order (int):
            The order of the grid. Defines the number of samples.
        dim (int):
            The number of dimensions in the grid

    Returns (numpy.ndarray):
        Regular grid with ``shape == (dim, order)``.
    """
    x_data = numpy.arange(1, order+1)/(order+1.)
    x_data = chaospy.quad.combine([x_data]*dim)
    return x_data.T


def create_nested_grid_samples(order, dim=1):
    """
    Create samples from a nested grid.

    Args:
        order (int):
            The order of the grid. Defines the number of samples.
        dim (int):
            The number of dimensions in the grid

    Returns (numpy.ndarray):
        Regular grid with ``shape == (dim, 2**order-1)``.
    """
    return create_grid_samples(order=2**order-1, dim=dim)



"""
A simple polynomial can be created through ``variable`` constructor. For
example to construct a simple bivariate polynomial::

     >>> x,y = cp.variable(dims=2)
     >>> print(x)
     q0

A collection of polynomial can be manipulated using basic arithmetic operators
and joined together into polynomial expansions::

     >>> polys = cp.Poly([1, x, x*y])
     >>> print(polys)
     [1, q0, q0q1]

Note that constants and simple polynomials can be joined together into arrays
without any problems.

In practice, having the ability to fine tune a polynomial exactly as one
wants it can be useful, but it can also be cumbersome when dealing with
larger arrays for application. To automate the construction of simple
polynomials, there is the ``basis`` constructor. In its simplest forms
it creates an array of simple monomials::

      >>> print(cp.basis(start=4))
      [1, q0, q0^2, q0^3, q0^4]

It can be expanded to include number of dimensions and a lower bound for the
polynomial order::

      >>> print(cp.basis(start=1, stop=2, dim=2))
      [q0, q1, q0^2, q0q1, q1^2]

There is also the possibility to create anisotropic expansions::

      >>> print(cp.basis(start=1, stop=[1, 2]))
      [q0, q1, q0q1, q1^2, q0q1^2]

and the possibility to fine tune the sorting of the polynomials::

      >>> print(cp.basis(start=1, stop=2, dim=2, sort="GRI"))
      [q0^2, q0q1, q1^2, q0, q1]

To manipulate a polynomial there is a collection of tools available. In
table [tab\_poly] they are listed with description. Much of the behavior
is overlapping with the numerical library ``numpy``. The functions that
overlaps in name is backwards compatible, e.g. if anything else than a
polynomial is inserted into ``cp.sum``, the argument is passed to
``numpy.sum``.

Any constructed polynomial is a callable. The argument can either be
inserted positional or as keyword arguments ``q0``, ``q1``, ...::

      >>> poly = cp.Poly([1, x**2, x*y])
      >>> print(poly(2, 3))
      [1 4 6]
      >>> print(poly(q1=3, q0=2))
      [1 4 6]

The input can be a mix of scalars and arrays, as long as the shapes together
can be joined to gether in a common compatible shape::

      >>> print(poly(2, [1, 2, 3, 4]))
      [[1 1 1 1]
       [4 4 4 4]
       [2 4 6 8]]

It is also possible to perform partial evaluation, i.e. evaluating some of the
dimensions. To tell the polynomial that a dimension should not be evaluated
either leave the argument empty or pass a masked value ``numpy.ma.masked``. For
example::

      >>> print(poly(2))
      [1, 4, 2q1]
      >>> print(poly(np.ma.masked, 2))
      [1, q0^2, 2q0]
      >>> print(poly(q1=2))
      [1, q0^2, 2q0]

The type of return value for the polynomial is ``numpy.ndarray`` if all
dimensions are filled. If not, it returns a new polynomial.

In addition to input scalars, arrays and masked values, it is also possible to
pass simple polynomials as argument. This allows for variable substitution. For
example, to swap two dimensions, one could simply do the following::

      >>> print(poly(y, x))
      [1, q1^2, q0q1]

It is also possible to do all the over mentioned methods together in the same
time. For example, partial evaluation and variable substitution::

      >>> print(poly(q1=y**3-1))
      [1, q0^2, q0q1^3-q0]
"""
import chaospy.poly.base
from chaospy.poly.base import Poly

import chaospy.poly.typing
import chaospy.poly.caller
import chaospy.poly.dimension
import chaospy.poly.shaping
import chaospy.poly.collection

from chaospy.poly.typing import *
from chaospy.poly.caller import (
    substitute, is_decomposed, decompose
)
from chaospy.poly.dimension import *
from chaospy.poly.shaping import *
from chaospy.poly.collection import *

import chaospy.poly.constructor



"""
Backend for the collection distributions.

To create a user-defined distribution use the distributions in
this module as template.

Documentation for each distribution is available in
distribution.collection.
"""
import numpy as np
import scipy as sp
from scipy import special

from .baseclass import Dist
from . import joint

class otdistribution(Dist):
    def __init__(self, distribution):
        Dist.__init__(self)
        if distribution.getDimension() != 1:
            raise Exception("Only 1D OpenTURNS distribution are supported for now")
        self.distribution = distribution
    def _pdf(self, x):
        return np.array(self.distribution.computePDF(np.atleast_2d(x).T).asPoint())
    def _cdf(self, x):
        return np.array(self.distribution.computeCDF(np.atleast_2d(x).T).asPoint())
    def _ppf(self, q):
        return np.array(self.distribution.computeQuantile(q[0]).asPoint())
    def _bnd(self):
        rng = self.distribution.getRange()
        return rng.getLowerBound()[0], rng.getUpperBound()[0]
    def _mom(self, k):
        return self.getMoment(k)[0]
    def _str(self):
        return self.distribution.__str__()


class uniform(Dist):

    def __init__(self):
        Dist.__init__(self)
    def _pdf(self, x):
        return 0.5
    def _cdf(self, x):
        return .5*x+.5
    def _ppf(self, q):
        return 2*q-1
    def _bnd(self):
        return -1.,1.
    def _mom(self, k):
        return 1./(k+1)*(k%2==0)
    def _ttr(self, n):
        return 0., n*n/(4.*n*n-1)
    def _str(self):
        return "uni"

class loguniform(Dist):

    def __init__(self, lo=0, up=1):
        Dist.__init__(self, lo=lo, up=up)
    def _pdf(self, x, lo, up):
        return 1./(x*(up-lo))
    def _cdf(self, x, lo, up):
        return (np.log(x)-lo)/(up-lo)
    def _ppf(self, q, lo, up):
        return np.e**(q*(up-lo) + lo)
    def _bnd(self, lo, up):
        return np.e**lo, np.e**up
    def _mom(self, k, lo, up):
        return ((np.e**(up*k)-np.e**(lo*k))/((up-lo)*(k+(k==0))))**(k!=0)
    def _str(self, lo, up):
        return "loguni(%s,%s)" % (lo, up)

class normal(Dist):

    def __init__(self):
        Dist.__init__(self)
    def _pdf(self, x):
        return (2*np.pi)**(-.5)*np.e**(-x**2/2.)
    def _cdf(self, x):
        return special.ndtr(x)
    def _ppf(self, x):
        return special.ndtri(x)
    def _mom(self, k):
        return .5*sp.misc.factorial2(k-1)*(1+(-1)**k)
    def _ttr(self, n):
        return 0., 1.*n
    def _bnd(self):
        return -7.5, 7.5
    def _str(self):
        return "nor"

class lognormal(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)
    def _pdf(self, x, a):
        out = np.e**(-np.log(x+(1-x)*(x<=0))**2/(2*a*a)) / \
            ((x+(1-x)*(x<=0))*a*np.sqrt(2*np.pi))*(x>0)
        return out
    def _cdf(self, x, a):
        return special.ndtr(np.log(x+(1-x)*(x<=0))/a)*(x>0)
    def _ppf(self, x, a):
        return np.e**(a*special.ndtri(x))
    def _mom(self, k, a):
        return np.e**(.5*a*a*k*k)
    def _ttr(self, n, a):
        return \
    (np.e**(n*a*a)*(np.e**(a*a)+1)-1)*np.e**(.5*(2*n-1)*a*a), \
                (np.e**(n*a*a)-1)*np.e**((3*n-2)*a*a)
    def _bnd(self, a):
        return 0, self._ppf(1-1e-10, a)
    def _str(self, a):
        return "lognor(%s)" % a


class expon(Dist):

    def __init__(self):
        Dist.__init__(self)
    def _pdf(self, x):
        return np.e**-x
    def _cdf(self, x):
        return 1.-np.e**-x
    def _ppf(self, q):
        return -np.log(1-q)
    def _mom(self, k):
        return sp.misc.factorial(k)
    def _ttr(self, n):
        return 2*n+1, n*n
    def _bnd(self):
        return 0, 42.
    def _str(self):
        return "expon"

class gamma(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)
    def _pdf(self, x, a):
        return x**(a-1)*np.e**(-x) / special.gamma(a)
    def _cdf(self, x, a):
        return special.gammainc(a, x)
    def _ppf(self, q, a):
        return special.gammaincinv(a, q)
    def _mom(self, k, a):
        return special.gamma(a+k)/special.gamma(a)
    def _ttr(self, n, a):
        return 2*n+a, n*n+n*(a-1)
    def _bnd(self, a):
        return 0, 40+2*a
    def _str(self, a):
        return "gam(%s)" % a

class laplace(Dist):

    def __init__(self):
        Dist.__init__(self)
    def _pdf(self, x):
        return np.e**-np.abs(x)/2
    def _cdf(self, x):
        return (1+np.sign(x)*(1-np.e**-abs(x)))/2
    def _mom(self, k):
        return .5*sp.misc.factorial(k)*(1+(-1)**k)
    def _ppf(self, x):
        return np.where(x>.5, -np.log(2*(1-x)), np.log(2*x))
    def _bnd(self):
        return -32., 32.
    def _str(self):
        return "lap"

class beta(Dist):

    def __init__(self, a=1, b=1):
        Dist.__init__(self, a=a, b=b)
    def _pdf(self, x, a, b):
        return x**(a-1)*(1-x)**(b-1)/ \
            special.beta(a, b)
    def _cdf(self, x, a, b):
        return special.btdtr(a, b, x)
    def _ppf(self, q, a, b):
        return special.btdtri(a, b, q)
    def _mom(self, k, a, b):
        return special.beta(a+k,b)/special.beta(a,b)
    def _ttr(self, n, a, b):

        nab = 2*n+a+b
        A = ((a-1)**2-(b-1)**2)*.5/\
                (nab*(nab-2) + (nab==0) + (nab==2)) + .5
        B1 = a*b*1./((a+b+1)*(a+b)**2)
        B2 = (n+a-1)*(n+b-1)*n*(n+a+b-2.)/\
            ((nab-1)*(nab-3)*(nab-2)**2+2.*((n==0)+(n==1)))
        B = np.where((n==0)+(n==1), B1, B2)
        return A, B
    def _bnd(self, a, b):
        return 0., 1.
    def _str(self, a, b):
        return "bet(%s,%s)" % (a,b)


class weibull(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)
    def _pdf(self, x, a):
        return a*x**(a-1)*np.e**(-x**a)
    def _cdf(self, x, a):
        return (1-np.e**(-x**a))
    def _ppf(self, q, a):
        return (-np.log(1-q+1*(q==1)))**(1./a)*(q!=1) +\
            30.**(1./a)*(q==1)
    def _mom(self, k, a):
        return special.gamma(1.+k*1./a)
    def _bnd(self, a):
        return 0, 30.**(1./a)
    def _str(self, a):
        return "wei(%s)" % a

def tri_ttr(k, a):
    from chaospy.quadrature import clenshaw_curtis
    q1,w1 = clenshaw_curtis(int(10**3*a), 0, a)
    q2,w2 = clenshaw_curtis(int(10**3*(1-a)), a, 1)
    q = np.concatenate([q1,q2], 1)
    w = np.concatenate([w1,w2])
    w = w*np.where(q<a, 2*q/a, 2*(1-q)/(1-a))

    from chaospy.poly import variable
    x = variable()

    orth = [x*0, x**0]
    inner = np.sum(q*w, -1)
    norms = [1., 1.]
    A,B = [],[]

    for n in range(k):
        A.append(inner/norms[-1])
        B.append(norms[-1]/norms[-2])
        orth.append((x-A[-1])*orth[-1]-orth[-2]*B[-1])

        y = orth[-1](*q)**2*w
        inner = np.sum(q*y, -1)
        norms.append(np.sum(y, -1))

    A, B = np.array(A).T[0], np.array(B).T
    return A, B


class triangle(Dist):

    def __init__(self, a=.5):
        assert np.all(a>=0) and np.all(a<=1)
        Dist.__init__(self, a=a)
    def _pdf(self, D, a):
        return np.where(D<a, 2*D/a, 2*(1-D)/(1-a))
    def _cdf(self, D, a):
        return np.where(D<a, D**2/(a + (a==0)),
                (2*D-D*D-a)/(1-a+(a==1)))
    def _ppf(self, q, a):
        return np.where(q<a, np.sqrt(q*a), 1-np.sqrt(1-a-q*(1-a)))
    def _mom(self, k, a):
        a_ = a*(a!=1)
        out = 2*(1.-a_**(k+1))/((k+1)*(k+2)*(1-a_))
        return np.where(a==1, 2./(k+2), out)
    def _bnd(self, a):
        return 0., 1.
    def _str(self, a):
        return "tri(%s)" % a
    def _ttr(self, k, a):
        a = a.item()
        if a==0: return beta()._ttr(k, 1, 2)
        if a==1: return beta()._ttr(k, 2, 1)

        A,B = tri_ttr(np.max(k)+1, a)
        A = np.array([[A[_] for _ in k[0]]])
        B = np.array([[B[_] for _ in k[0]]])
        return A,B


#  class wigner(Dist):
#
#      def __init__(self):
#          Dist.__init__(self)
#
#      def _pdf(self, x):
#          return 2*np.sqrt(1-x*x)/np.pi
#      def _cdf(self, x):
#          return special.btdtr(1.5, 1.5, .5*x+.5)
#      def _ppf(self, q):
#          return 2*special.btdtri(1.5, 1.5, q)-1
#      def _mom(self, n):
#          return ((n+1)%2)*comb(n, n/2)/((.5*n+1)*2**n)
#      def _bnd(self):
#          return -1.,1.
#      def _ttr(self, n):
#          return 0., .25**(n!=0)
#      def _str(self):
#          return "w"


class kumaraswamy(Dist):

    def __init__(self, a=1, b=1):
        assert np.all(a>0) and np.all(b>0)
        Dist.__init__(self, a=a, b=b)

    def _pdf(self, x, a, b):
        return a*b*x**(a-1)*(1-x**a)**(b-1)

    def _cdf(self, x, a, b):
        return 1-(1-x**a)**b

    def _ppf(self, q, a, b):
        return (1-(1-q)**(1./b))**(1./a)

    def _mom(self, k, a, b):
        return b*special.gamma(1+k*1./a)*special.gamma(b)/\
                special.gamma(1+b+k*1./a)

    def _str(self, a, b):
        return "kum(%s,%s)" % (a,b)

    def _bnd(self, a, b):
        return 0,1

class hypgeosec(Dist):

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return .5*np.cosh(np.pi*x/2.)**-1

    def _cdf(self, x):
        return 2/np.pi*np.arctan(np.e**(np.pi*x/2.))

    def _ppf(self, q):
        return 2/np.pi*np.log(np.tan(np.pi*q/2.))

    def _mom(self, k):
        return np.abs(special.euler(k))[-1]

    def _str(self):
        return "hgs"

class logistic(Dist):

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return np.e**-x/(1+np.e**-x)**(c+1)

    def _cdf(self, x, c):
        return (1+np.e**-x)**-c

    def _ppf(self, q, c):
        return -np.log(q**(-1/c)-1)

    def _bnd(self, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)

    def _str(self, c):
        return "log(%s)" % c

class student_t(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        return special.gamma(.5*a+.5)*(1+x*x/a)**(-.5*a-.5) /\
                (np.sqrt(a*np.pi)*special.gamma(.5*a))

    def _cdf(self, x, a):
        return special.stdtr(a, x)

    def _ppf(self, q, a):
        return special.stdtrit(a, q)

    def _bnd(self, a):
        return self._ppf(1e-10, a), self._ppf(1-1e-10, a)

    def _mom(self, k, a):
        if np.any(a<=k):
            raise ValueError("too high mom for student-t")
        out = special.gamma(.5*k+.5)* \
                special.gamma(.5*a-.5*k)*a**(.5*k)
        return np.where(k%2==0, out/(np.pi**.5*special.gamma(.5*a)), 0)

    def _ttr(self, k, a):
        return 0., k*a*(a-k+1.)/ ((a-2*k)*(a-2*k+2))

    def _str(self, a):
        return "stt(%s)" % a


class raised_cosine(Dist):

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return .5 + .5*np.cos(np.pi*x)

    def _cdf(self, x):
        return .5 + .5*x + np.sin(np.pi*x)/(2*np.pi)

    def _bnd(self):
        return -1,1

    def _mom(self, k):
        return np.where(k%2, 0, 2/(k+2) + 1/(k+1)*\
                special.hyp1f2((k+1)/2.), .5, (k+3)/2., -np.pi**2/4)

    def _str(self):
        return "cos"

class mvnormal(Dist):

    def __init__(self, loc=[0,0], scale=[[1,.5],[.5,1]]):
        loc, scale = np.asfarray(loc), np.asfarray(scale)
        C = np.linalg.cholesky(scale)
        Ci = np.linalg.inv(C)
        Dist.__init__(self, C=C, Ci=Ci, loc=loc,
                _advance=True, _length=len(C))

    def _cdf(self, x, graph):
        Ci, loc = graph.keys["Ci"], graph.keys["loc"]
        return sp.special.ndtr(np.dot(Ci, (x.T-loc.T).T))

    def _ppf(self, q, graph):
        return (np.dot(graph.keys["C"], sp.special.ndtri(q)).T+graph.keys["loc"].T).T

    def _pdf(self, x, graph):

        loc, C, Ci = graph.keys["loc"], graph.keys["C"], graph.keys["Ci"]
        det = np.linalg.det(np.dot(C,C.T))

        x_ = np.dot(Ci.T, (x.T-loc.T).T)
        out = np.ones(x.shape)
        out[0] =  np.e**(-.5*np.sum(x_*x_, 0))/np.sqrt((2*np.pi)**len(Ci)*det)
        return out

    def _bnd(self, x, graph):

        C, loc = graph.keys["C"], graph.keys["loc"]
        scale = np.sqrt(np.diag(np.dot(C,C.T)))
        lo,up = np.zeros((2,)+x.shape)
        lo.T[:] = (-7.5*scale+loc)
        up.T[:] = (7.5*scale+loc)
        return lo,up

    def _mom(self, k, graph):

        C, loc = graph.keys["C"], graph.keys["loc"]
        scale = np.dot(C, C.T)

        def mom(k):

            zeros = (np.sum(k,0)%2==1)+np.any(np.array(k)<0, 0)
            if np.all(zeros, 0):
                return 0.

            dim, K = k.shape
            ra = np.arange(dim).repeat(K).reshape(dim,K)

            i = np.argmax(k!=0, 0)

            out = np.zeros(k.shape[1:])
            out[:] = np.where(np.choose(i,k),
                    (np.choose(i,k)-1)*scale[i,i]*mom(k-2*(ra==i)), 1)
            for x in range(1, dim):
                out += \
                (np.choose(i,k)!=0)*(x>i)*k[x]*scale[i,x]*mom(k-(ra==i)-(ra==x))

            return out

        dim = len(loc)
        K = np.mgrid[[slice(0,_+1,1) for _ in np.max(k, 1)]]
        K = K.reshape(dim, int(K.size/dim))
        M = mom(K)

        out = np.zeros(k.shape[1])
        for i in range(len(M)):
            coef = np.prod(sp.misc.comb(k.T, K[:,i]).T, 0)
            diff = k.T - K[:,i]
            pos = diff>=0
            diff = diff*pos
            pos = np.all(pos, 1)
            loc_ = np.prod(loc**diff, 1)
            out += pos*coef*loc_*M[i]

        return out

    def _dep(self, graph):
        n = normal()
        out = [set([n]) for _ in range(len(self))]
        return out

    def _str(self, C, loc, **prm):
        return "mvnor(%s,%s)" % (loc, C)

class mvlognormal(Dist):

    def __init__(self, loc=[0,0], scale=[[1,.5],[.5,1]]):

        loc, scale = np.asfarray(loc), np.asfarray(scale)
        assert len(loc)==len(scale)

        dist = joint.Iid(normal(), len(loc))
        C = np.linalg.cholesky(scale)
        Ci = np.linalg.inv(C)
        Dist.__init__(self, dist=dist, loc=loc, C=C, Ci=Ci,
                scale=scale, _length=len(scale), _advance=True)

    def _cdf(self, x, graph):

        y = np.log(np.abs(x) + 1.*(x<=0))
        out = graph(np.dot(graph.keys["Ci"], (y.T-graph.keys["loc"].T).T),
                graph.dists["dist"])
        return np.where(x<=0, 0., out)

    def _ppf(self, q, graph):
        return np.e**(np.dot(graph.keys["C"], \
                graph(q, graph.dists["dist"])).T+graph.keys["loc"].T).T

    def _mom(self, k, graph):
        scale, loc = graph.keys["scale"], graph.keys["loc"]
        return np.e**(np.dot(k.T, loc).T+ \
            .5*np.diag(np.dot(k.T, np.dot(scale, k))))

    def _bnd(self, x, graph):
        loc, scale = graph.keys["loc"], graph.keys["scale"]
        up = (7.1*np.sqrt(np.diag(scale))*x.T**0 + loc.T).T
        return 0*up, np.e**up

    def _val(self, graph):
        if "dist" in graph.keys:
            return (np.dot(graph.keys["dist"].T, graph.keys["C"].T)+graph.keys["loc"].T).T
        return self

    def _dep(self, graph):

        dist = graph.dists["dist"]
        S = graph(dist)
        out = [set([]) for _ in range(len(self))]
        C = graph.keys["C"]

        for i in range(len(self)):
            for j in range(len(self)):
                if C[i,j]:
                    out[i].update(S[j])
        return out

    def _str(self, loc, C, **prm):
        print("mvlognor(%s,%s)" % (loc, C))

class mvstudentt(Dist):

    def __init__(self, a=1, loc=[0,0], scale=[[1,.5],[.5,1]]):
        loc, scale = np.asfarray(loc), np.asfarray(scale)
        C = np.linalg.cholesky(scale)
        Ci = np.linalg.inv(C)
        Dist.__init__(self, a=a, C=C, Ci=Ci, loc=loc, _length=len(C))

    def _cdf(self, x, a, C, Ci, loc):
        x = np.dot(Ci, (x.T-loc.T).T)
        return special.stdtr(a, x)

    def _ppf(self, q, a, C, Ci, loc):
        z = sp.special.stdtrit(a, q)
        out = (np.dot(C, z).T + loc.T).T
        return out

    def _pdf(self, x, a, C, Ci, loc):

        det = np.linalg.det(np.dot(C,C.T))
        k = len(C)

        x_ = np.dot(Ci.T, (x.T-loc.T).T)
        out = np.ones(x.shape)
        out[0] = special.gamma(.5*(a+k))/(special.gamma(.5*a)* \
                a**(.5*k)*np.pi**(.5*k)*det**.5*\
                (1+np.sum(x_*x_,0)/a))
        return out

    def _bnd(self, a, C, Ci, loc):

        scale = np.sqrt(np.diag(np.dot(C,C.T)))
        lo,up = np.zeros((2,len(self)))
        lo.T[:] = (-10**5*scale+loc)
        up.T[:] = (10**5*scale+loc)
        return lo,up

#      def _mom(self, k, graph):
#
#          C, loc = graph.keys["C"], graph.keys["loc"]
#          scale = np.dot(C, C.T)
#
#          def mom(k):
#
#              zeros = (np.sum(k,0)%2==1)+np.any(np.array(k)<0, 0)
#              if np.all(zeros, 0):
#                  return 0.
#
#              dim, K = k.shape
#              ra = np.arange(dim).repeat(K).reshape(dim,K)
#
#              i = np.argmax(k!=0, 0)
#
#              out = np.zeros(k.shape[1:])
#              out[:] = np.where(np.choose(i,k),
#                      (np.choose(i,k)-1)*scale[i,i]*mom(k-2*(ra==i)), 1)
#              for x in range(1, dim):
#                  out += \
#                  (np.choose(i,k)!=0)*(x>i)*k[x]*scale[i,x]*mom(k-(ra==i)-(ra==x))
#
#              return out
#
#          dim = len(loc)
#          K = np.mgrid[[slice(0,_+1,1) for _ in np.max(k, 1)]]
#          K = K.reshape(dim, K.size/dim)
#          M = mom(K)
#
#          out = np.zeros(k.shape[1])
#          for i in range(len(M)):
#              coef = np.prod(sp.misc.comb(k.T, K[:,i]).T, 0)
#              diff = k.T - K[:,i]
#              pos = diff>=0
#              diff = diff*pos
#              pos = np.all(pos, 1)
#              loc_ = np.prod(loc**diff, 1)
#              out += pos*coef*loc_*M[i]
#
#          return out

    def _dep(self, graph):
        n = student_t()
        out = [set([n]) for _ in range(len(self))]
        return out

    def _str(self, a, loc, C, **prm):
        return "mvstt(%s,%s,%s)" % (a,loc,C)

#  class Dirichlet(be.Dist):
#      """
#  Dirichlet \sim Dir(alpha)
#
#  Parameters
#  ----------
#  alpha : array_like
#      Shape parameters.
#      len(alpha)>1
#      np.all(alpha>0)
#
#  Examples
#  --------
#  >>> chaospy.seed(1000)
#  >>> f = chaospy.Dirichlet([1,2,3])
#  >>> q = [[.3,.3,.7,.7],[.3,.7,.3,.7]]
#  >>> print(f.inv(q))
#  [[ 0.06885008  0.06885008  0.21399691  0.21399691]
#   [ 0.25363028  0.47340104  0.21409462  0.39960771]]
#  >>> print(f.fwd(f.inv(q)))
#  [[ 0.3  0.3  0.7  0.7]
#   [ 0.3  0.7  0.3  0.7]]
#  >>> print(f.sample(4))
#  [[ 0.12507651  0.00904026  0.06508353  0.07888277]
#   [ 0.29474152  0.26985323  0.69375006  0.30848838]]
#  >>> print(f.mom((1,1)))
#  0.047619047619
#      """
#
#      def __init__(self, alpha=[1,1,1]):
#
#          dists = [co.beta() for _ in range(len(alpha)-1)]
#          ba.Dist.__init__(self, _dists=dists, alpha=alpha, _name="D")
#
#      def _upd(self, alpha, **prm):
#
#          alpha = alpha.flatten()
#          dim = len(alpha)-1
#          out = [None]*dim
#          _dists = prm.pop("_" + self.name)
#          cum = _dists[0]
#
#          _dists[0].upd(a=alpha[0], b=np.sum(alpha[1:], 0))
#          out[0] = _dists[0]
#          for i in range(1, dim):
#              _dists[i].upd(a=alpha[i], b=np.sum(alpha[i+1:], 0))
#              out[i] = _dists[i]*(1-cum)
#              cum = cum+out[i]
#
#          prm = dict(alpha=alpha)
#          prm["_" + self.name] = out
#          return prm
#
#      def _mom(self, k, alpha, **prm):
#
#          out = np.empty(k.shape[1:])
#          out[:] = sp.special.gamma(np.sum(alpha, 0))
#          out /= sp.special.gamma(np.sum(alpha, 0)+np.sum(k, 0))
#          out *= np.prod(sp.special.gamma(alpha[:-1]+k.T).T, 0)
#          out /= np.prod(sp.special.gamma(alpha[:-1]), 0)
#          return out
#


##NEW

class alpha(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _cdf(self, x, a):
        return special.ndtr(a-1./x) / special.ndtr(a)

    def _ppf(self, q, a):
        return 1.0/(a-special.ndtri(q*special.ndtr(a)))

    def _pdf(self, x, a):
        return 1.0/(x**2)/special.ndtr(a)*np.e**(.5*(a-1.0/x)**2)/np.sqrt(2*np.pi)

    def _bnd(self, a):
        return 0,self._ppf(1-1e-10, a)

class anglit(Dist):

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return np.cos(2*x)
    def _cdf(self, x):
        return np.sin(x+np.pi/4)**2.0
    def _ppf(self, q):
        return (np.arcsin(np.sqrt(q))-np.pi/4)
    def _bnd(self):
        return -np.pi/4, np.pi/4


class bradford(Dist):

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return  c / (c*x + 1.0) / np.log(1.0+c)
    def _cdf(self, x, c):
        return np.log(1.0+c*x) / np.log(c+1.0)
    def _ppf(self, q, c):
        return ((1.0+c)**q-1)/c
    def _bnd(self, c):
        return 0, 1

class burr(Dist):

    def __init__(self, c=1., d=1.):
        Dist.__init__(self, c=c, d=d)
    def _pdf(self, x, c, d):
        return c*d*(x**(-c-1.0))*((1+x**(-c*1.0))**(-d-1.0))
    def _cdf(self, x, c, d):
        return (1+x**(-c*1.0))**(-d**1.0)
    def _ppf(self, q, c, d):
        return (q**(-1.0/d)-1)**(-1.0/c)
    def _bnd(self, c, d):
        return 0, self._ppf(1-1e-10, c, d)
    def _mom(self, k, c, d):
        return d*special.beta(1-k*1./c, d+k*1./c)

class fisk(Dist):

    def __init__(self, c=1.):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        return c*(x**(-c-1.0))*((1+x**(-c*1.0))**(-1.0))
    def _cdf(self, x, c):
        return (1+x**(-c*1.0))**(-1.0)
    def _ppf(self, q, c):
        return (q**(-1.0)-1)**(-1.0/c)
    def _bnd(self, c):
        return 0, self._ppf(1-1e-10, c)

class cauchy(Dist):

    def __init__(self):
        Dist.__init__(self)
    def _pdf(self, x):
        return 1.0/np.pi/(1.0+x*x)
    def _cdf(self, x):
        return 0.5 + 1.0/np.pi*np.arctan(x)
    def _ppf(self, q):
        return np.tan(np.pi*q-np.pi/2.0)
    def _bnd(self):
        return self._ppf(1e-10), self._ppf(1-1e-10)

class chi(Dist):

    def __init__(self, df=1):
        Dist.__init__(self, df=df)
    def _pdf(self, x, df):
        return x**(df-1.)*np.exp(-x*x*0.5)/(2.0)**(df*0.5-1)\
                /special.gamma(df*0.5)
    def _cdf(self, x, df):
        return special.gammainc(df*0.5,0.5*x*x)
    def _ppf(self, q, df):
        return np.sqrt(2*special.gammaincinv(df*0.5,q))
    def _bnd(self, df):
        return 0, self._ppf(1-1e-10, df)
    def _mom(self, k, df):
        return 2**(.5*k)*special.gamma(.5*(df+k))\
                /special.gamma(.5*df)

class dbl_gamma(Dist):

    def __init__(self, a):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        ax = abs(x)
        return 1.0/(2*special.gamma(a))*ax**(a-1.0) * np.exp(-ax)

    def _cdf(self, x, a):
        fac = 0.5*special.gammainc(a,abs(x))
        return np.where(x>0,0.5+fac,0.5-fac)

    def _ppf(self, q, a):
        fac = special.gammainccinv(a,1-abs(2*q-1))
        return np.where(q>0.5, fac, -fac)

    def _bnd(self, a):
        return self._ppf(1e-10, a), self._ppf(1-1e-10, a)

class dbl_weibull(Dist):

    def __init__(self, c):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        ax = np.abs(x)
        Px = c/2.0*ax**(c-1.0)*np.exp(-ax**c)
        return Px
    def _cdf(self, x, c):
        Cx1 = 0.5*np.exp(-abs(x)**c)
        return np.where(x > 0, 1-Cx1, Cx1)
    def _ppf(self, q, c):
        q_ = np.where(q>.5, 1-q, q)
        Cq1 = (-np.log(2*q_))**(1./c)
        return np.where(q>.5, Cq1, -Cq1)
    def _bnd(self, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)


class erlang(Dist):


    def __init__(self, a=1):
        Dist.__init__(self, a=a)
    def _pdf(self, x, a):
        Px = (x)**(a-1.0)*np.exp(-x)/special.gamma(a)
        return Px
    def _cdf(self, x, a):
        return special.gdtr(1.0,a,x)
    def _ppf(self, q, a):
        return special.gdtrix(1.0, a, q)
    def _bnd(self, a):
        return 0, self._ppf(1-1e-10, a)

class exponweibull(Dist):

    def __init__(self, a=1, c=1):
        Dist.__init__(self, a=a, c=c)
    def _pdf(self, x, a, c):
        exc = np.exp(-x**c)
        return a*c*(1-exc)**(a-1) * exc * x**(c-1)
    def _cdf(self, x, a, c):
        exm1c = -np.expm1(-x**c)
        return (exm1c)**a
    def _ppf(self, q, a, c):
        return (-np.log1p(-q**(1.0/a)))**(1.0/c)
    def _bnd(self, a, c):
        return 0, self._ppf(1-1e-10, a, c)

class exponpow(Dist):

    def __init__(self, b=1):
        Dist.__init__(self, b=b)
    def _pdf(self, x, b):
        xbm1 = x**(b-1.0)
        xb = xbm1 * x
        return np.exp(1)*b*xbm1 * np.exp(xb - np.exp(xb))
    def _cdf(self, x, b):
        xb = x**b
        return -np.expm1(-np.expm1(xb))
    def _ppf(self, q, b):
        return pow(np.log1p(-np.log1p(-q)), 1.0/b)
    def _bnd(self, b):
        return 0,self._ppf(1-1e-10, b)

class fatiguelife(Dist):

    def __init__(self, c=0):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        return (x+1)/(2*c*np.sqrt(2*np.pi*x**3))*np.exp(-(x-1)**2/(2.0*x*c**2))
    def _cdf(self, x, c):
        return special.ndtr(1.0/c*(np.sqrt(x)-1.0/np.sqrt(x)))
    def _ppf(self, q, c):
        tmp = c*special.ndtri(q)
        return 0.25*(tmp + np.sqrt(tmp**2 + 4))**2
    def _bnd(self, c):
        return 0, self._ppf(1-1e-10, c)

class foldcauchy(Dist):

    def __init__(self, c=0):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        return 1.0/np.pi*(1.0/(1+(x-c)**2) + 1.0/(1+(x+c)**2))
    def _cdf(self, x, c):
        return 1.0/np.pi*(np.arctan(x-c) + np.arctan(x+c))
    def _bnd(self, c):
        return 0, 10**10


class foldnorm(Dist):

    def __init__(self, c=1):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        return np.sqrt(2.0/np.pi)*np.cosh(c*x)*np.exp(-(x*x+c*c)/2.0)
    def _cdf(self, x, c):
        return special.ndtr(x-c) + special.ndtr(x+c) - 1.0
    def _bnd(self, c):
        return 0, 7.5+c

class frechet(Dist):
    def __init__(self, c=1):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        return c*pow(x,c-1)*np.exp(-pow(x,c))
    def _cdf(self, x, c):
        return -np.expm1(-pow(x,c))
    def _ppf(self, q, c):
        return pow(-np.log1p(-q),1.0/c)
    def _mom(self, k, c):
        return special.gamma(1-k*1./c)
    def _bnd(self, c):
        return 0, self._ppf(1-1e-10, c)


class genexpon(Dist):
    def __init__(self, a=1, b=1, c=1):
        Dist.__init__(self, a=a, b=b, c=c)
    def _pdf(self, x, a, b, c):
        return (a+b*(-np.expm1(-c*x)))*np.exp((-a-b)*x+b*(-np.expm1(-c*x))/c)
    def _cdf(self, x, a, b, c):
        return -np.expm1((-a-b)*x + b*(-np.expm1(-c*x))/c)
    def _bnd(self, a, b, c):
        return 0, 10**10

class genextreme(Dist):
    def __init__(self, c=1):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        cx = c*x
        logex2 = np.where((c==0)*(x==x),0.0,np.log1p(-cx))
        logpex2 = np.where((c==0)*(x==x),-x,logex2/c)
        pex2 = np.exp(logpex2)
        logpdf = np.where((cx==1) | (cx==-np.inf),-np.inf,-pex2+logpex2-logex2)
        np.putmask(logpdf,(c==1) & (x==1),0.0)
        return np.exp(logpdf)

    def _cdf(self, x, c):
        loglogcdf = np.where((c==0)*(x==x),-x,np.log1p(-c*x)/c)
        return np.exp(-np.exp(loglogcdf))

    def _ppf(self, q, c):
        x = -np.log(-np.log(q))
        return np.where((c==0)*(x==x),x,-np.expm1(-c*x)/c)
    def _bnd(self, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)


class gengamma(Dist):

    def __init__(self, x, a, c):
        Dist.__init__(self, a=a, c=c)
    def _pdf(self, x, a, c):
        return abs(c)* np.exp((c*a-1)*np.log(x)-x**c- special.gammaln(a))
    def _cdf(self, x, a, c):
        val = special.gammainc(a,x**c)
        cond = c + 0*val
        return np.where(cond>0,val,1-val)
    def _ppf(self, q, a, c):
        val1 = special.gammaincinv(a,q)
        val2 = special.gammaincinv(a,1.0-q)
        ic = 1.0/c
        cond = c+0*val1
        return np.where(cond > 0,val1**ic,val2**ic)
    def _mom(self, k, a, c):
        return special.gamma((c+k)*1./a)/special.gamma(c*1./a)
    def _bnd(self, a, c):
        return 0.0, self._ppf(1-1e-10, a, c)


class genhalflogistic(Dist):

    def __init__(self, c=1):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        limit = 1.0/c
        tmp = (1-c*x)
        tmp0 = tmp**(limit-1)
        tmp2 = tmp0*tmp
        return 2*tmp0 / (1+tmp2)**2
    def _cdf(self, x, c):
        limit = 1.0/c
        tmp = (1-c*x)
        tmp2 = tmp**(limit)
        return (1.0-tmp2) / (1+tmp2)
    def _ppf(self, q, c):
        return 1.0/c*(1-((1.0-q)/(1.0+q))**c)
    def _bnd(self, c):
        return 0.0, 1/np.where(c<10**-10, 10**-10, c)


class gompertz(Dist):

    def __init__(self, c):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        ex = np.exp(x)
        return c*ex*np.exp(-c*(ex-1))
    def _cdf(self, x, c):
        return 1.0-np.exp(-c*(np.exp(x)-1))
    def _ppf(self, q, c):
        return np.log(1-1.0/c*np.log(1-q))
    def _bnd(self, c):
        return 0.0, self._ppf(1-1e-10, c)


class gumbel(Dist):

    def __init__(self):
        Dist.__init__(self)
    def _pdf(self, x):
        ex = np.exp(-x)
        return ex*np.exp(-ex)
    def _cdf(self, x):
        return np.exp(-np.exp(-x))
    def _ppf(self, q):
        return -np.log(-np.log(q))
    def _bnd(self):
        return self._ppf(1e-10), self._ppf(1-1e-10)


class levy(Dist):

    def __init__(self):
        Dist.__init__(self)
    def _pdf(self, x):
        return 1/np.sqrt(2*np.pi*x)/x*np.exp(-1/(2*x))
    def _cdf(self, x):
        return 2*(1-normal._cdf(1/np.sqrt(x)))
    def _ppf(self, q):
        val = normal._ppf(1-q/2.0)
        return 1.0/(val*val)
    def _bnd(self):
        return 0.0, self._ppf(1-1e-10)


class loggamma(Dist):

    def __init__(self, c):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        return np.exp(c*x-np.exp(x)-special.gammaln(c))
    def _cdf(self, x, c):
        return special.gammainc(c, np.exp(x))
    def _ppf(self, q, c):
        return np.log(special.gammaincinv(c,q))
    def _bnd(self, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)


class loglaplace(Dist):

    def __init__(self, c):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        cd2 = c/2.0
        c = np.where(x < 1, c, -c)
        return cd2*x**(c-1)
    def _cdf(self, x, c):
        return np.where(x < 1, 0.5*x**c, 1-0.5*x**(-c))
    def _ppf(self, q, c):
        return np.where(q < 0.5, (2.0*q)**(1.0/c), (2*(1.0-q))**(-1.0/c))
    def _bnd(self, c):
        return 0.0, self._ppf(1-1e-10, c)



class mielke(Dist):

    def __init__(self, k, s):
        Dist.__init__(self, k=k, s=s)
    def _pdf(self, x, k, s):
        return k*x**(k-1.0) / (1.0+x**s)**(1.0+k*1.0/s)
    def _cdf(self, x, k, s):
        return x**k / (1.0+x**s)**(k*1.0/s)
    def _ppf(self, q, k, s):
        qsk = pow(q,s*1.0/k)
        return pow(qsk/(1.0-qsk),1.0/s)
    def _bnd(self, k, s):
        return 0.0, self._ppf(1-1e-10, k, s)


class nakagami(Dist):
    def __init__(self, nu):
        Dist.__init__(self, nu=nu)
    def _pdf(self, x, nu):
        return 2*nu**nu/special.gamma(nu)*(x**(2*nu-1.0))*np.exp(-nu*x*x)
    def _cdf(self, x, nu):
        return special.gammainc(nu,nu*x*x)
    def _ppf(self, q, nu):
        return np.sqrt(1.0/nu*special.gammaincinv(nu,q))
    def _bnd(self, nu):
        return 0.0, self._ppf(1-1e-10)


class chisquared(Dist):

    def __init__(self, df, nc):
        Dist.__init__(self, df=df, nc=nc)
    def _pdf(self, x, df, nc):
        a = df/2.0
        fac = (-nc-x)/2.0 + (a-1)*np.log(x)-a*np.log(2)-special.gammaln(a)
        fac += np.nan_to_num(np.log(special.hyp0f1(a, nc * x/4.0)))
        return np.np.exp(fac)
    def _cdf(self, x, df, nc):
        return special.chndtr(x,df,nc)
    def _ppf(self, q, df, nc):
        return special.chndtrix(q,df,nc)
    def _bnd(self, df, nc):
        return 0.0, self._ppf(1-1e-10, nc)

#  class f(Dist):
#
#      def __init__(self, n=1, m=1):
#          Dist.__init__(self, n=n, m=m)
#      def _pdf(self, x, n, m):
#          lPx = m/2*np.np.log(m) + n/2*np.log(n) + (n/2-1)*np.log(x)
#          lPx -= ((n+m)/2)*np.log(m+n*x) + special.betaln(n/2,m/2)
#          return np.exp(lPx)
#      def _cdf(self, x, n, m):
#          return special.fdtr(n, m, x)
#      def _ppf(self, q, n, m):
#          return special.fdtri(n, m, q)
#      def _mom(self, k, n, m):
#          ga = special.gamma
#          return (n*1./m)**k*ga(.5*n+k)*ga(.5*m-k)/ga(.5*n)/ga(.5*m)
#      def _bnd(self, n, m):
#          return 0, self._ppf(1-1e-10, n, m)

class f(Dist):
    """A non-central F distribution continuous random variable.

    %(before_notes)s

    Notes
    -----
    The probability density function for `ncf` is::

    ncf.pdf(x, df1, df2, nc) = np.exp(nc/2 + nc*df1*x/(2*(df1*x+df2)))
                    * df1**(df1/2) * df2**(df2/2) * x**(df1/2-1)
                    * (df2+df1*x)**(-(df1+df2)/2)
                    * gamma(df1/2)*gamma(1+df2/2)
                    * L^{v1/2-1}^{v2/2}(-nc*v1*x/(2*(v1*x+v2)))
                    / (B(v1/2, v2/2) * gamma((v1+v2)/2))

    for ``df1, df2, nc > 0``.

    %(example)s

    """
    def __init__(self, dfn, dfd, nc):
        Dist.__init__(self, dfn=dfn, dfd=dfd, nc=nc)
    def _cdf(self, x, dfn, dfd, nc):
        return special.ncfdtr(dfn,dfd,nc,x)
    def _ppf(self, q, dfn, dfd, nc):
        return special.ncfdtri(dfn, dfd, nc, q)
    def _bnd(self, dfn, dfd, nc):
        return 0.0, self._ppf(1-1e-10, dfn, dfd, nc)


class nct(Dist):

    def __init__(self, df, nc):
        Dist.__init__(self, df=df, nc=nc)
    def _pdf(self, x, df, nc):
        n = df*1.0
        nc = nc*1.0
        x2 = x*x
        ncx2 = nc*nc*x2
        fac1 = n + x2
        trm1 = n/2.*np.log(n) + special.gammaln(n+1)
        trm1 -= n*np.log(2)+nc*nc/2.+(n/2.)*np.log(fac1)+special.gammaln(n/2.)
        Px = np.exp(trm1)
        valF = ncx2 / (2*fac1)
        trm1 = np.sqrt(2)*nc*x*special.hyp1f1(n/2+1,1.5,valF)
        trm1 /= (fac1*special.gamma((n+1)/2))
        trm2 = special.hyp1f1((n+1)/2,0.5,valF)
        trm2 /= (np.sqrt(fac1)*special.gamma(n/2+1))
        Px *= trm1+trm2
        return Px
    def _cdf(self, x, df, nc):
        return special.nctdtr(df, nc, x)
    def _ppf(self, q, df, nc):
        return special.nctdtrit(df, nc, q)
    def _bnd(self, df, nc):
        return self._ppf(1e-10, df, nc), self._ppf(1-1e-10, df, nc)


#  class pareto(Dist):
#      def __init__(self, c=1):
#          Dist.__init__(self, c=c)
#      def _pdf(self, x, c):
#          Px = pow(1+c*x,-1.-1./c)
#          return Px
#      def _cdf(self, x, c):
#          return 1.0 - pow(1+c*x,-1.0/c)
#      def _ppf(self, q, c):
#          vals = 1.0/c * (pow(1-q, -c)-1)
#          return vals
#      def _bnd(self, c):
#          return 1, self._ppf(1-1e-10, c)

class pareto1(Dist):
    def __init__(self, b):
        Dist.__init__(self, b=b)
    def _pdf(self, x, b):
        return b * x**(-b-1)
    def _cdf(self, x, b):
        return 1 -  x**(-b)
    def _ppf(self, q, b):
        return pow(1-q, -1.0/b)
    def _bnd(self, b):
        return 1.0, self._ppf(1-1e-10, b)
class pareto2(Dist):

    def _pdf(self, x, c):
        return c*1.0/(1.0+x)**(c+1.0)
    def _cdf(self, x, c):
        return 1.0-1.0/(1.0+x)**c
    def _ppf(self, q, c):
        return pow(1.0-q,-1.0/c)-1
    def _bnd(self, c):
        return 0.0, self._ppf(1-1e-10, c)


class powerlognorm(normal):

    def __init__(self, c, s):
        Dist.__init__(self, c=c, s=s)
    def _pdf(self, x, c, s):
        return c/(x*s)*normal._pdf(self, \
                np.log(x)/s)*pow(normal._cdf(self, -np.log(x)/s),c*1.0-1.0)

    def _cdf(self, x, c, s):
        return 1.0 - pow(normal._cdf(self, -np.log(x)/s),c*1.0)
    def _ppf(self, q, c, s):
        return np.exp(-s*normal._ppf(self, pow(1.0-q,1.0/c)))
    def _bnd(self, c, s):
        return 0.0, self._ppf(1-1e-10, c, s)


class powernorm(Dist):

    def __init__(self, c):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        return c*normal._pdf(x)* \
               (normal._cdf(-x)**(c-1.0))
    def _cdf(self, x, c):
        return 1.0-normal._cdf(-x)**(c*1.0)
    def _ppf(self, q, c):
        return -normal._ppf(pow(1.0-q,1.0/c))
    def _bnd(self, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)

class wald(Dist):

    def __init__(self, mu):
        Dist.__init__(self, mu=mu)
    def _pdf(self, x, mu):
        return 1.0/np.sqrt(2*np.pi*x)*np.exp(-(1-mu*x)**2.0 / (2*x*mu**2.0))
    def _cdf(self, x, mu):
        trm1 = 1.0/mu - x
        trm2 = 1.0/mu + x
        isqx = 1.0/np.sqrt(x)
        return 1.0-normal._cdf(self, isqx*trm1)-\
                np.exp(2.0/mu)*normal._cdf(self, -isqx*trm2)
    def _bnd(self, mu):
        return 0.0, 10**10

class reciprocal(Dist):

    def __init__(self, lo=0, up=1):
        Dist.__init__(self, lo=lo, up=up)
    def _pdf(self, x, lo, up):
        return 1./(x*np.log(up/lo))
    def _cdf(self, x, lo, up):
        return np.log(x/lo)/np.log(up/lo)
    def _ppf(self, q, lo, up):
        return np.e**(q*np.log(up/lo) + np.log(lo))
    def _bnd(self, lo, up):
        return lo, up
    def _mom(self, k, lo, up):
        return ((up*np.e**k-lo*np.e**k)/(np.log(up/lo)*(k+(k==0))))**(k!=0)

class truncexpon(Dist):

    def __init__(self, b):
        Dist.__init__(self, b=b)
    def _pdf(self, x, b):
        return np.exp(-x)/(1-np.exp(-b))
    def _cdf(self, x, b):
        return (1.0-np.exp(-x))/(1-np.exp(-b))
    def _ppf(self, q, b):
        return -np.log(1-q+q*np.exp(-b))
    def _bnd(self, b):
        return 0.0, b


class truncnorm(Dist):

    def __init__(self, a, b, mu, sigma):
        Dist.__init__(self, a=a, b=b)
        self.norm = normal()*sigma+mu
        self.fa = self.norm.fwd(a)
        self.fb = self.norm.fwd(b)
    def _pdf(self, x, a, b):
        return self.norm.pdf(x) / (self.fb-self.fa)
    def _cdf(self, x, a, b):
        return (self.norm.fwd(x) - self.fa) / (self.fb-self.fa)
    def _ppf(self, q, a, b):
        return self.norm.inv(q*(self.fb-self.fa) + self.fa)
    def _bnd(self, a, b):
        return a, b


class tukeylambda(Dist):

    def __init__(self, lam):
        Dist.__init__(self, lam=lam)
    def _pdf(self, x, lam):
        Fx = (special.tklmbda(x,lam))
        Px = Fx**(lam-1.0) + ((1-Fx))**(lam-1.0)
        Px = 1.0/(Px)
        return np.where((lam <= 0) | (abs(x) < 1.0/(lam)), Px, 0.0)

    def _cdf(self, x, lam):
        return special.tklmbda(x, lam)

    def _ppf(self, q, lam):
        q = q*1.0
        vals1 = (q**lam - (1-q)**lam)/lam
        vals2 = np.log(q/(1-q))
        return np.where((lam==0)&(q==q), vals2, vals1)

    def _bnd(self, lam):
        return self._ppf(1e-10, lam), self._ppf(1-1e-10, lam)



class wrapcauchy(Dist):

    def __init__(self, c):
        Dist.__init__(self, c=c)
    def _pdf(self, x, c):
        return (1.0-c*c)/(2*np.pi*(1+c*c-2*c*np.cos(x)))
    def _cdf(self, x, c):
        output = 0.0*x
        val = (1.0+c)/(1.0-c)
        c1 = x<np.pi
        c2 = 1-c1

        xn = np.extract(c2,x)
        if (any(xn)):
            valn = np.extract(c2, np.ones_like(x)*val)
            xn = 2*np.pi - xn
            yn = np.tan(xn/2.0)
            on = 1.0-1.0/np.pi*np.arctan(valn*yn)
            np.place(output, c2, on)

        xp = np.extract(c1,x)
        if (any(xp)):
            valp = np.extract(c1, np.ones_like(x)*val)
            yp = np.tan(xp/2.0)
            op = 1.0/np.pi*np.arctan(valp*yp)
            np.place(output, c1, op)

        return output

    def _ppf(self, q, c):
        val = (1.0-c)/(1.0+c)
        rcq = 2*np.arctan(val*np.tan(np.pi*q))
        rcmq = 2*np.pi-2*np.arctan(val*np.tan(np.pi*(1-q)))
        return np.where(q < 1.0/2, rcq, rcmq)
    def _bnd(self, c):
        return 0.0, 2*np.pi

class rice(Dist):

    def __init__(self, a):
        Dist.__init__(a=a)

    def _pdf(self, x, a):
        return x*np.exp(-.5*(x*x+a*a))*special.j0(x*a)

    def _cdf(self, x, a):
        return special.chndtr(x*x, 2, a*a)

    def _ppf(self, q, a):
        return special.chdtrix(np.sqrt(q), 2, a*a)

    def _bnd(self, a):
        return 0, special.chndtrix(np.sqrt(1-1e-10), 2, a*a)

class kdedist(Dist):
    """
A distribution that is based on a kernel density estimator (KDE).
    """
    def __init__(self, kernel, lo, up):
        self.kernel = kernel
        super(kdedist, self).__init__(lo=lo, up=up)

    def _cdf(self, x, lo, up):
        cdf_vals = np.zeros(x.shape)
        for i in range(0, len(x)):
            cdf_vals[i] = [self.kernel.integrate_box_1d(0, x_i) for x_i in x[i]]
        return cdf_vals

    def _pdf(self, x, lo, up):
        return self.kernel(x)

    def _bnd(self, lo, up):
        return (lo, up)

    def sample(self, size=(), rule="R", antithetic=None,
            verbose=False, **kws):
        """
            Overwrite sample() function, because the constructed Dist that is
            based on the KDE is only working with the random sampling that is
            given by the KDE itself.
        """

        size_ = np.prod(size, dtype=int)
        dim = len(self)
        if dim>1:
            if isinstance(size, (tuple,list,np.ndarray)):
                shape = (dim,) + tuple(size)
            else:
                shape = (dim, size)
        else:
            shape = size

        out = self.kernel.resample(size_)[0]
        try:
            out = out.reshape(shape)
        except:
            if len(self)==1:
                out = out.flatten()
            else:
                out = out.reshape(dim, out.size/dim)

        return out



# pylint: disable=protected-access
"""Backend for value callback function."""
import numpy


def val_call(self, dist):
    """Value callback wrapper."""
    graph = self.graph

    if "key" in graph.node[dist]:
        return graph.node[dist]["key"]

    self.dist, dist_ = dist, self.dist
    for value in dist.prm.values():
        if not isinstance(value, numpy.ndarray) and\
                not "key" in graph.node[value]:
            val_call(self, value)

    if hasattr(dist, "_val"):
        out = dist._val(self)

        if isinstance(out, numpy.ndarray):
            graph.add_node(dist, key=out)

    else:
        out = dist

    self.dist = dist_
    return out



"""
Functions mimicing linear algebra operations.
"""

import numpy as np
import chaospy as cp

import chaospy.poly
from chaospy.poly.base import Poly


def inner(*args):
    """
    Inner product of a polynomial set.

    Args:
        arg0, arg1, [...] (Poly) : The polynomials to perform inner product on.

    Returns:
        (Poly) : Resulting polynomial.

    Examples:
        >>> x,y = cp.variable(2)
        >>> P = cp.Poly([x-1, y])
        >>> Q = cp.Poly([x+1, x*y])
        >>> print(cp.inner(P, Q))
        q0^2+q0q1^2-1
        >>> x = np.arange(4)
        >>> print(cp.inner(x, x))
        14
    """
    haspoly = sum([isinstance(arg, Poly) for arg in args])

    # Numpy
    if not haspoly:
        return np.sum(np.prod(args, 0), 0)

    # Poly
    out = args[0]
    for arg in args[1:]:
        out = out * arg
    return sum(out)


def outer(*args):
    """
    Polynomial outer product.

    Args:
        P1 (Poly, ndarray, int, float) : First term in outer product
        P2 (Poly, array_like) : Second term in outer product

    Returns:
        (Poly) : Poly set with same dimensions as itter.

    Examples:
        >>> x = cp.variable()
        >>> P = cp.prange(3)
        >>> print(P)
        [1, q0, q0^2]
        >>> print(cp.outer(x, P))
        [q0, q0^2, q0^3]
        >>> print(cp.outer(P, P))
        [[1, q0, q0^2], [q0, q0^2, q0^3], [q0^2, q0^3, q0^4]]
    """
    if len(args) > 2:
        part1 = args[0]
        part2 = outer(*args[1:])

    elif len(args) == 2:
        part1, part2 = args

    else:
        return args[0]

    dtype = chaospy.poly.typing.dtyping(part1, part2)

    if dtype in (list, tuple, np.ndarray):

        part1 = np.array(part1)
        part2 = np.array(part2)
        shape = part1.shape +  part2.shape
        return np.outer(
            chaospy.poly.shaping.flatten(part1),
            chaospy.poly.shaping.flatten(part2),
        )

    if dtype == Poly:

        if isinstance(part1, Poly) and isinstance(part2, Poly):

            if (1,) in (part1.shape, part2.shape):
                return part1*part2

            shape = part1.shape+part2.shape

            out = []
            for _ in chaospy.poly.shaping.flatten(part1):
                out.append(part2*_)

            return chaospy.poly.shaping.reshape(Poly(out), shape)

        if isinstance(part1, (int, float, list, tuple)):
            part2, part1 = np.array(part1), part2

        else:
            part2 = np.array(part2)

        core_old = part1.A
        core_new = {}
        for key in part1.keys:
            core_new[key] = outer(core_old[key], part2)
        shape = part1.shape+part2.shape
        dtype = chaospy.poly.typing.dtyping(part1.dtype, part2.dtype)
        return Poly(core_new, part1.dim, shape, dtype)

    raise NotImplementedError


def dot(poly1, poly2):
    """
    Dot product of polynomial vectors.

    Args:
        poly1 (Poly) : left part of product.
        poly2 (Poly) : right part of product.

    Returns:
        (Poly) : product of poly1 and poly2.

    Examples:
        >>> poly = cp.prange(3, 1)
        >>> print(poly)
        [1, q0, q0^2]
        >>> print(cp.dot(poly, np.arange(3)))
        2q0^2+q0
        >>> print(cp.dot(poly, poly))
        q0^4+q0^2+1
    """
    if not isinstance(poly1, Poly) and not isinstance(poly2, Poly):
        return np.dot(poly1, poly2)

    poly1 = Poly(poly1)
    poly2 = Poly(poly2)

    poly = poly1*poly2
    if np.prod(poly1.shape) <= 1 or np.prod(poly2.shape) <= 1:
        return poly
    return chaospy.poly.sum(poly, 0)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Collection of trigonometric operators
"""
import numpy as np

from ..baseclass import Dist

class Sin(Dist):

    def __init__(self, dist):
        lo,up = dist.range()
        assert np.all(lo>=-np.pi/2) and np.all(up<=np.pi/2)
        Dist.__init__(self, dist=dist, _length=len(dist),
                _advance=True)

    def _str(self, dist):
        return "Sin(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return np.sin(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(np.arcsin(x), graph.dists["dist"])/np.sqrt(1-x*x)

    def _cdf(self, x, graph):
        return graph(np.arcsin(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return np.sin(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return np.sin(graph(np.arcsin(x), graph.dists["dist"]))

class Arcsin(Dist):

    def __init__(self, dist):
        lo,up = dist.range()
        assert np.all(lo>=-1) and np.all(up<=1)
        Dist.__init__(self, dist=dist, _length=len(dist),
                _advance=True)

    def _str(self, dist):
        return "Arcsin(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return np.arcsin(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(np.sin(x), graph.dists["dist"])*np.cos(x)

    def _cdf(self, x, graph):
        return graph(np.sin(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return np.arcsin(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return np.arcsin(graph(np.sin(x), graph.dists["dist"]))


class Cos(Dist):

    def __init__(self, dist):
        lo,up = dist.range()
        assert np.all(lo>=0) and np.all(up<=np.pi)
        Dist.__init__(self, dist=dist, _length=len(dist),
                _advance=True)

    def _str(self, dist):
        return "Cos(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return np.cos(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(np.arccos(x), graph.dists["dist"])/np.sqrt(1-x*x)

    def _cdf(self, x, graph):
        return 1-graph(np.arccos(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return np.cos(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return np.cos(graph(np.arccos(x), graph.dists["dist"]))



class Arccos(Dist):

    def __init__(self, dist):
        lo,up = dist.range()
        assert np.all(lo>=-1) and np.all(up<=1)
        Dist.__init__(self, dist=dist, _length=len(dist),
                _advance=True)

    def _str(self, dist):
        return "Arccos(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return np.arccos(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(np.cos(x), graph.dists["dist"])*np.sin(x)

    def _cdf(self, x, graph):
        return 1-graph(np.cos(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return np.arccos(graph(-q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return np.arccos(graph(np.cos(x), graph.dists["dist"]))

class Tan(Dist):

    def __init__(self, dist):
        lo,up = dist.range()
        assert np.all(lo>=-np.pi/2) and np.all(up<=np.pi/2)
        Dist.__init__(self, dist=dist, _length=len(dist),
                _advance=True)

    def _str(self, dist):
        return "Tan(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return np.tan(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(np.arctan(x), graph.dists["dist"])/(1-x*x)

    def _cdf(self, x, graph):
        return graph(np.arctan(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return np.tan(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return np.tan(graph(np.arctan(x), graph.dists["dist"]))

class Arctan(Dist):

    def __init__(self, dist):
        Dist.__init__(self, dist=dist, _length=len(dist),
                _advance=True)

    def _str(self, dist):
        return "Arccos(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return np.arctan(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(np.tan(x), graph.dists["dist"])/np.cos(x)**2

    def _cdf(self, x, graph):
        return graph(np.cos(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return np.arctan(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return np.arctan(graph(np.tan(x), graph.dists["dist"]))



"""
Point collcation method, or regression based polynomial chaos expansion builds
open the idea of fitting a polynomial chaos expansion to a set of generated
samples and evaluations. The experiement can be done as follows:

- Select a :ref:`distributions`::

    >>> distribution = cp.Iid(cp.Normal(0, 1), 2)

- Generate :ref:`orthogonality`::

    >>> orthogonal_expansion = cp.orth_ttr(2, distribution)
    >>> print(orthogonal_expansion)
    [1.0, q1, q0, q1^2-1.0, q0q1, q0^2-1.0]

- Generate samples using :ref:`montecarlo` (or alternative absissas from
  :ref:`quadrature`)::

    >>> samples = distribution.sample(
    ...     2*len(orthogonal_expansion), rule="M")
    >>> print(samples[:, :4])
    [[ 0.67448975 -1.15034938  0.31863936 -0.31863936]
     [-1.42607687 -1.02007623 -0.73631592 -0.50240222]]

- A function evaluated using the nodes generated in the second step::

    >>> def model_solver(param):
    ...     return [param[0]*param[1], param[0]*np.e**-param[1]+1]
    >>> solves = [model_solver(sample) for sample in samples.T]
    >>> print(np.around(solves[:4], 8))
    [[-0.96187423  3.80745414]
     [ 1.17344406 -2.19038608]
     [-0.23461924  1.66539168]
     [ 0.16008512  0.47338898]]

- Bring it all together using `~chaospy.collocation.fit_regression`::

    >>> approx_model = cp.fit_regression(
    ...      orthogonal_expansion, samples, solves)
    >>> print(cp.around(approx_model, 5))
    [q0q1, 0.0478q0^2-1.4354q0q1+0.1108q1^2+1.22377q0-0.0907q1+0.93973]

In this example, the number of collocation points is selected to be twice the
number of unknown coefficients :math:`N+1`. This
follows the default outlined in :cite:`hosder_efficient_2007`. Changing this is
obviously possible. When the number of parameter is equal the number of
unknown, the, the polynomial approximation becomes an interpolation method and
overlap with Lagrange polynomials. If the number of samples are fewer than the
number of unknown, classical least squares can not be used.  Instead it
possible to use methods for doing estimation with too few samples.

The function :func:`~chaospy.collocation.fit_regression` also takes an optional
``rule`` keyword argument. It allows for the selection of regression method
used when fitting the samples to the polynomials:

+------------+----------------------------------------------+
| Key        | Description                                  |
| Parameters |                                              |
+============+==============================================+
| "LS"       | Ordenary Least Squares                       |
+------------+----------------------------------------------+
| "T"        | Ridge Regression/Tikhonov Regularization     |
| order      | Order of regularization (or custom matrix)   |
| alpha      | Dampning parameter (else estimated from gcv) |
+------------+----------------------------------------------+
| "TC"       | Tikhonov with Cross Validation               |
| order      | Order of regularization (or custom matrix)   |
| alpha      | Dampning parameter (else estimated from gcv) |
+------------+----------------------------------------------+

In addition, the rule can be any initialized regression model from sklearn.
For example, if one wants to implement orthogonal matching pursuit
:cite:`mallat_matching_1993`, it can for example be implemented as follows::
. It forces the
result to have at most one non-zero coefficient. To implement it use the
keyword ``rule="OMP"``, and to force the number of coefficients to be
for example 1: ``n_nonzero_coefs=1``. In practice::

    >>> from sklearn.linear_model import OrthogonalMatchingPursuit
    >>> omp = OrthogonalMatchingPursuit(fit_intercept=False, n_nonzero_coefs=1)
    >>> approx_model = cp.fit_regression(
    ...     orthogonal_expansion, samples, solves, rule=omp)
    >>> print(cp.around(approx_model, 8))
    [3.46375077q0q1, 11.63750715]

Note that the option `fit_intercept=False`. This is a prerequisite for
``sklearn`` to be compatible with ``chaospy``.
"""
import numpy as np
from scipy import linalg

import chaospy as cp

__all__ = ("fit_regression", "rlstsq")


def fit_regression(
        polynomials,
        abscissas,
        evals,
        rule="LS",
        retall=False,
        order=0,
        alpha=None,
):
    """
    Fit a polynomial chaos expansion using linear regression.

    Args:
        polynomials (Poly):
            Polynomial expansion with `polynomials.shape=(M,)` and
            `polynomials.dim=D`.
        abscissas (array_like):
            Collocation nodes with `abscissas.shape=(D,K)`.
        evals (array_like):
            Model evaluations with `len(evals)=K`.
        retall (bool):
            If True return Fourier coefficients in addition to R.
        order (int):
            Tikhonov regularization order.
        alpha (float):
            Dampning parameter for the Tikhonov regularization. Calculated
            automatically if omitted.

    Returns:
        (Poly, np.ndarray):
            Fitted polynomial with `R.shape=evals.shape[1:]` and `R.dim=D`. The
            Fourier coefficients in the estimation.

    Examples:
        >>> x, y = cp.variable(2)
        >>> polynomials = cp.Poly([1, x, y])
        >>> abscissas = [[-1,-1,1,1], [-1,1,-1,1]]
        >>> evals = [0,1,1,2]
        >>> print(cp.around(cp.fit_regression(polynomials, abscissas, evals), 14))
        0.5q0+0.5q1+1.0
    """
    abscissas = np.asarray(abscissas)
    if len(abscissas.shape) == 1:
        abscissas = abscissas.reshape(1, *abscissas.shape)
    evals = np.array(evals)

    poly_evals = polynomials(*abscissas).T
    shape = evals.shape[1:]
    evals = evals.reshape(evals.shape[0], int(np.prod(evals.shape[1:])))

    if isinstance(rule, str):
        rule = rule.upper()

    if rule == "LS":
        uhat = linalg.lstsq(poly_evals, evals)[0]

    elif rule == "T":
        uhat = rlstsq(poly_evals, evals, order=order, alpha=alpha, cross=False)

    elif rule == "TC":
        uhat = rlstsq(poly_evals, evals, order=order, alpha=alpha, cross=True)

    else:

        from sklearn.linear_model.base import LinearModel
        assert isinstance(rule, LinearModel)
        uhat = rule.fit(poly_evals, evals).coef_.T

    evals = evals.reshape(evals.shape[0], *shape)

    approx_model = cp.poly.sum((polynomials*uhat.T), -1)
    approx_model = cp.poly.reshape(approx_model, shape)

    if retall == 1:
        return approx_model, uhat
    elif retall == 2:
        return approx_model, uhat, poly_evals
    return approx_model


def rlstsq(coef_mat, ordinate, order=1, alpha=None, cross=False):
    """
    Least Squares Minimization using Tikhonov regularization.

    Includes method for robust generalized cross-validation.

    Args:
        coef_mat (array_like):
            Coefficient matrix with shape (M,N).
        ordinate (array_like, shape (M,) or (M, K)):
            Ordinate or "dependent variable" values with shape (M,) or (M, K).
            If `ordinate` is two-dimensional, the least-squares solution is
            calculated for each of the `K` columns of `ordinate`.
        order (int, array_like):
            If int, it is the order of Tikhonov regularization.  If array_like,
            it will be used as regularization matrix.
        alpha (float, optional):
            Lower threshold for the dampening parameter. The real value is
            calculated using generalised cross validation.
        cross (bool, optional):
            Use cross validation to estimate alpha value.
    """
    coef_mat = np.array(coef_mat)
    ordinate = np.array(ordinate)
    dim1, dim2 = coef_mat.shape

    if cross:
        out = np.empty((dim1, dim2) + ordinate.shape[1:])
        coef_mat_ = np.empty((dim1-1, dim2))
        ordinate_ = np.empty((dim1-1,) + ordinate.shape[1:])
        for i in range(dim1):
            coef_mat_[:i] = coef_mat[:i]
            coef_mat_[i:] = coef_mat[i+1:]
            ordinate_[:i] = ordinate[:i]
            ordinate_[i:] = ordinate[i+1:]
            out[i] = rlstsq(coef_mat_, ordinate_, order, alpha, False)

        return np.median(out, 0)

    if order == 0:
        tikhmat = np.eye(dim2)

    elif order == 1:
        tikhmat = np.zeros((dim2-1, dim2))
        tikhmat[:, :-1] -= np.eye(dim2-1)
        tikhmat[:, 1:] += np.eye(dim2-1)

    elif order == 2:
        tikhmat = np.zeros((dim2-2, dim2))
        tikhmat[:, :-2] += np.eye(dim2-2)
        tikhmat[:, 1:-1] -= 2*np.eye(dim2-2)
        tikhmat[:, 2:] += np.eye(dim2-2)

    elif order is None:
        tikhmat = np.zeros(1)

    else:
        tikhmat = np.array(order)
        assert tikhmat.shape[-1] == dim2 or tikhmat.shape in ((), (1,))

    if alpha is None and order is not None:

        gamma = 0.1

        def rgcv_error(alpha):
            """Calculate Tikhonov dampening parameter."""
            if alpha <= 0:
                return np.inf
            coef_mat_ = np.dot(
                coef_mat.T, coef_mat)+alpha*(np.dot(tikhmat.T, tikhmat))
            try:
                coef_mat_ = np.dot(linalg.inv(coef_mat_), coef_mat.T)
            except linalg.LinAlgError:
                return np.inf

            abscissas = np.dot(coef_mat_, ordinate)
            res2 = np.sum((np.dot(coef_mat, abscissas)-ordinate)**2)
            coef_mat_2 = np.dot(coef_mat, coef_mat_)
            skew = dim1*res2/np.trace(np.eye(dim1)-coef_mat_2)**2
            mu2 = np.sum(coef_mat_2*coef_mat_2.T)/dim1

            return (gamma + (1-gamma)*mu2)*skew

        alphas = 10.**-np.arange(0, 16)
        evals = np.array([rgcv_error(alpha) for alpha in alphas])
        alpha = alphas[np.argmin(evals)]

    out = linalg.inv(
        np.dot(coef_mat.T, coef_mat) + alpha*np.dot(tikhmat.T, tikhmat))
    out = np.dot(out, np.dot(coef_mat.T, ordinate))
    return out


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Addition operator.

Example usage
-------------

Distribution and a constant::

    >>> distribution = chaospy.Normal(0, 1) + 10
    >>> print(distribution)
    (Normal(0,1)+10)
    >>> print(numpy.around(distribution.sample(5), 4))
    [10.395   8.7997 11.6476  9.9553 11.1382]
    >>> print(numpy.around(distribution.fwd([9, 10, 11]), 4))
    [0.1587 0.5    0.8413]
    >>> print(numpy.around(distribution.inv(distribution.fwd([9, 10, 11])), 4))
    [ 9. 10. 11.]
    >>> print(numpy.around(distribution.pdf([9, 10, 11]), 4))
    [0.242  0.3989 0.242 ]
    >>> print(distribution.mom([1, 2, 3]))
    [  10.  101. 1030.]
    >>> print(distribution.ttr([1, 2, 3]))
    [[10. 10. 10.]
     [ 1.  2.  3.]]

Construct joint addition distribution::

    >>> lhs = chaospy.Uniform(2, 3)
    >>> rhs = chaospy.Uniform(3, 4)
    >>> addition = lhs + rhs
    >>> print(addition)
    (Uniform(2,3)+Uniform(3,4))
    >>> joint1 = chaospy.J(lhs, addition)
    >>> joint2 = chaospy.J(rhs, addition)

Generate random samples::

    >>> print(numpy.around(joint1.sample(4), 4))
    [[2.2123 2.0407 2.3972 2.2331]
     [6.0541 5.2478 6.1397 5.6253]]
    >>> print(numpy.around(joint2.sample(4), 4))
    [[3.1823 3.7435 3.0696 3.8853]
     [6.1349 6.6747 5.485  5.9143]]

Forward transformations::

    >>> lcorr = numpy.array([2.1, 2.5, 2.9])
    >>> rcorr = numpy.array([3.01, 3.5, 3.99])
    >>> print(numpy.around(joint1.fwd([lcorr, lcorr+rcorr]), 4))
    [[0.1  0.5  0.9 ]
     [0.01 0.5  0.99]]
    >>> print(numpy.around(joint2.fwd([rcorr, lcorr+rcorr]), 4))
    [[0.01 0.5  0.99]
     [0.1  0.5  0.9 ]]

Inverse transformations::

    >>> print(numpy.around(joint1.inv(joint1.fwd([lcorr, lcorr+rcorr])), 4))
    [[2.1  2.5  2.9 ]
     [5.11 6.   6.89]]
    >>> print(numpy.around(joint2.inv(joint2.fwd([rcorr, lcorr+rcorr])), 4))
    [[3.01 3.5  3.99]
     [5.11 6.   6.89]]

Raw moments::

    >>> print(numpy.around(joint1.mom([(0, 1, 1), (1, 0, 1)]), 4))
    [ 6.   2.5 15. ]
    >>> print(numpy.around(joint2.mom([(0, 1, 1), (1, 0, 1)]), 4))
    [ 6.   3.5 21. ]
"""
from scipy.misc import comb
import numpy

from ..baseclass import Dist
from .multiply import mul


class Add(Dist):
    """Addition."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, array_like) : Left hand side.
            right (Dist, array_like) : Right hand side.
        """
        try:
            left_ = len(left)
        except TypeError:
            left_ = 1
        except IndexError:
            left_ = 1
        try:
            right_ = len(right)
        except TypeError:
            right_ = 1
        except IndexError:
            right_ = 1
        length = max(left_, right_)
        Dist.__init__(self, left=left, right=right, _length=length, _advance=True)

    def _str(self, left, right):
        """String representation."""
        return "(%s+%s)" % (left, right)

    def _val(self, graph):
        """Value extraction."""
        if len(graph.keys) == 2:
            return graph.keys["left"] + graph.keys["right"]
        return self

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        lower, upper = graph((xloc.T).T, dist)
        return (lower.T+num.T).T, (upper.T+num.T).T

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        return graph((xloc.T-num.T).T, dist)

    def _pdf(self, xloc, graph):
        """Probability density function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        return graph((xloc.T-num.T).T, dist)

    def _ppf(self, uloc, graph):
        """Point percentile function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        return (graph(uloc, dist).T + num.T).T

    def _mom(self, keys, graph):
        """Statistical moments."""
        if len(graph.dists) == 2 and \
                graph.dists["left"].dependent(graph.dists["right"]):
            raise NotImplementedError("dependency")

        dim = len(self)
        kmax = numpy.max(keys, 1)+1
        keys_ = numpy.mgrid[[slice(0, _, 1) for _ in kmax]]
        keys_ = keys_.reshape(dim, int(keys_.size/dim))

        left = []
        if "left" in graph.dists:
            left.append(graph(keys_, graph.dists["left"]))
        else:
            left.append(graph.keys["left"]**keys_.T)
        if "right" in graph.dists:
            left.append(graph(keys_, graph.dists["right"]))
        else:
            left.append(graph.keys["right"]**keys_.T)

        if dim == 1:
            left[0] = left[0].flatten()
            left[1] = left[1].flatten()
            keys = keys.flatten()

        out = 0.
        for idx in range(keys_.shape[1]):
            key = keys_.T[idx]
            coef = comb(keys.T, key)
            out = out + coef*left[0][idx]*left[1][keys-idx]*(key <= keys.T)

        if dim > 1:
            out = numpy.prod(out, 1)
        return out

    def _ttr(self, order, graph):
        """Three terms recursion coefficients."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]

        coeffs0, coeffs1 = graph(order, dist)
        return coeffs0+num, coeffs1


def add(left, right):
    """
    Distribution addition.

    Args:
        left (Dist, array_like) : left hand side.
        right (Dist, array_like) : right hand side.
    """
    if left is right:
        return mul(2, left)

    if isinstance(left, Dist):
        if not isinstance(right, Dist):
            right = numpy.array(right)
            if right.size == 1 and right == 0:
                return left

    elif isinstance(right, Dist):
        left = numpy.array(left)
        if left.size == 1 and left == 0:
            return right

    else:
        return left+right

    return Add(left=left, right=right)



"""
Gauss-Legendre quadrature rule.
"""
import numpy

import chaospy.quad


def quad_gauss_legendre(order, lower=0, upper=1, composite=None):
    """
    Generate the quadrature nodes and weights in Gauss-Legendre quadrature.

    Example:
        >>> abscissas, weights = quad_gauss_legendre(3)
        >>> print(numpy.around(abscissas, 4))
        [[0.0694 0.33   0.67   0.9306]]
        >>> print(numpy.around(weights, 4))
        [0.1739 0.3261 0.3261 0.1739]
    """
    order = numpy.asarray(order, dtype=int).flatten()
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)
    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    if composite is None:
        composite = numpy.array(0)
    composite = numpy.asarray(composite)

    if not composite.size:
        composite = numpy.array([numpy.linspace(0, 1, composite+1)]*dim)

    else:
        composite = numpy.array(composite)
        if len(composite.shape) <= 1:
            composite = numpy.transpose([composite])
        composite = ((composite.T-lower)/(upper-lower)).T

    results = [_gauss_legendre(order[i], composite[i]) for i in range(dim)]
    abscis = numpy.array([_[0] for _ in results])
    weights = numpy.array([_[1] for _ in results])

    abscis = chaospy.quad.combine(abscis)
    weights = chaospy.quad.combine(weights)

    abscis = (upper-lower)*abscis + lower
    weights = numpy.prod(weights*(upper-lower), 1)

    return abscis.T, weights


def _gauss_legendre(order, composite=1):
    """Backend function."""
    inner = numpy.ones(order+1)*0.5
    outer = numpy.arange(order+1)**2
    outer = outer/(16*outer-4.)

    banded = numpy.diag(numpy.sqrt(outer[1:]), k=-1) + numpy.diag(inner) + \
            numpy.diag(numpy.sqrt(outer[1:]), k=1)
    vals, vecs = numpy.linalg.eig(banded)

    abscis, weight = vals.real, vecs[0, :]**2
    indices = numpy.argsort(abscis)
    abscis, weight = abscis[indices], weight[indices]

    n_abscis = len(abscis)

    composite = numpy.array(composite).flatten()
    composite = list(set(composite))
    composite = [comp for comp in composite if (comp < 1) and (comp > 0)]
    composite.sort()
    composite = [0]+composite+[1]

    abscissas = numpy.empty(n_abscis*(len(composite)-1))
    weights = numpy.empty(n_abscis*(len(composite)-1))
    for dim in range(len(composite)-1):
        abscissas[dim*n_abscis:(dim+1)*n_abscis] = \
            abscis*(composite[dim+1]-composite[dim]) + composite[dim]
        weights[dim*n_abscis:(dim+1)*n_abscis] = \
            weight*(composite[dim+1]-composite[dim])

    return abscissas, weights



"""
Methods for performing Gaussian quadrature.

Examples
--------
A quadrature rule where the limits are defined by floats::

    >>> abscissas, weights = chaospy.quad_clenshaw_curtis(
    ...         order=3, lower=0., upper=1.)
    >>> print(numpy.around(abscissas, 4))
    [[0.   0.25 0.75 1.  ]]
    >>> print(numpy.around(weights, 4))
    [0.1111 0.3889 0.1944 0.1111]

A quadrature rule in higher dimensions through the main interface::

    >>> lower = [0., 0.]
    >>> upper = [1., 1.]
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...         order=1, domain=(lower, upper), rule="gauss_legendre")
    >>> print(numpy.around(abscissas, 4))
    [[0.2113 0.2113 0.7887 0.7887]
     [0.2113 0.7887 0.2113 0.7887]]
    >>> print(numpy.around(weights, 4))
    [0.25 0.25 0.25 0.25]

A quadrature rule where the distribution is the weight function::

    >>> dist = chaospy.Gamma()
    >>> abscissas, weights = chaospy.quad_golub_welsch(
    ...         order=2, dist=dist)
    >>> print(numpy.around(abscissas, 4))
    [[0.4158 2.2943 6.2899]]
    >>> print(numpy.around(weights, 4))
    [0.7111 0.2785 0.0104]
"""
from .combine import combine
from .interface import generate_quadrature

from .stieltjes import generate_stieltjes
from .collection import *
from .sparse_grid import sparse_grid
from .generator import rule_generator



"""
Trunkation of distribution.
"""
import numpy

from ..baseclass import Dist


class Trunk(Dist):
    """Trunkation."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, array_like) : Left hand side.
            right (Dist, array_like) : Right hand side.
        """
        if isinstance(left, Dist):
            assert not left.dependent()
            assert numpy.all(left.range()[0] < right)
        if isinstance(right, Dist):
            assert not right.dependent()
            assert numpy.all(left < right.range()[1])

        left_ = not isinstance(left, Dist) or 1 and len(left)
        right_ = not isinstance(right, Dist) or 1 and len(right)
        length = max(left_, right_)

        Dist.__init__(
            self, left=left, right=right, _length=length, _advance=True)

    def _str(self, left, right):
        """String representation."""
        return "(%s<%s)" % (left, right)

    def _pdf(self, xloc, graph):
        """
        Probability density function.

        Example:
            >>> dist = chaospy.trunk(chaospy.Uniform(), 0.6)
            >>> print(dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25]))
            [0.         1.66666667 1.66666667 0.         0.        ]
            >>> dist = chaospy.trunk(chaospy.Uniform(), 0.4)
            >>> print(dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25]))
            [0.  2.5 0.  0.  0. ]
            >>> dist = chaospy.trunk(0.4, chaospy.Uniform())
            >>> print(dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25]))
            [0.         0.         1.66666667 1.66666667 0.        ]
            >>> dist = chaospy.trunk(0.6, chaospy.Uniform())
            >>> print(dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25]))
            [0.  0.  0.  2.5 0. ]
        """
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
            num = (num.T*numpy.ones(xloc.shape[::-1])).T
            out = graph(xloc, dist)/(1-dist.fwd(num))
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
            num = (num.T*numpy.ones(xloc.shape[::-1])).T
            out = graph(xloc, dist)/dist.fwd(num)

        return out

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        graph_ = graph.copy()
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
            num = (num.T*numpy.ones(xloc.shape[::-1])).T
            uloc1 = graph(xloc, dist)
            uloc2 = graph_(num, dist)
            out = (uloc1-uloc2)/(1-uloc2)
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
            num = (num.T*numpy.ones(xloc.shape[::-1])).T
            uloc1 = graph(xloc, dist)
            uloc2 = graph_(num, dist)
            out = (uloc1)/uloc2

        return out

    def _ppf(self, q, graph):
        """Point percentile function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
            num = (num.T*numpy.ones(q.shape[::-1])).T
            uloc = dist.fwd(num)
            out = graph(q*(1-uloc)+uloc, dist)
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
            num = (num.T*numpy.ones(q.shape[::-1])).T
            uloc = dist.fwd(num)
            out = graph(q*uloc, dist)

        return out

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
            num = (num.T*numpy.ones(xloc.shape[::-1])).T
            xloc = numpy.where(xloc > num, num, xloc)
            lower, upper = graph(xloc, dist)
            lower = numpy.max([num, lower], 0)
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
            num = (num.T*numpy.ones(xloc.shape[::-1])).T
            xloc = numpy.where(xloc < num, num, xloc)
            lower, upper = graph(xloc, dist)
            upper = numpy.min([num, upper], 0)

        return lower, upper

def trunk(left, right):
    """
    Trunkation.

    Args:
        left (Dist, array_like) : left hand side.
        right (Dist, array_like) : right hand side.
    """
    if not isinstance(left, Dist) and not isinstance(right, Dist):
        return left < right
    return Trunk(left, right)



"""
Power operator.
"""
import numpy

from ..baseclass import Dist


class Pow(Dist):
    """Power operator."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, array_like) : Left hand side.
            right (Dist, array_like) : Right hand side.
        """
        left_ = 1 if not isinstance(left, Dist) else len(left)
        right_ = 1 if not isinstance(right, Dist) else len(right)
        length = max(left_,right_)
        Dist.__init__(self, left=left, right=right,
                _length=length, _advance=True)

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        if "left" in graph.keys and "right" in graph.dists:

            num, dist = graph.keys["left"], graph.dists["right"]

            assert numpy.all(num>=0), "root of negative number"

            y = numpy.where(xloc<0, -numpy.inf,
                    numpy.log(numpy.where(xloc>0, xloc, 1)))/\
                        numpy.log(num*(1.-(num==1)))

            bnd = num**graph(y, dist)
            correct = bnd[0]<bnd[1]
            bnd_ = numpy.empty(bnd.shape)
            bnd_[0] = numpy.where(correct, bnd[0], bnd[1])
            bnd_[1] = numpy.where(correct, bnd[1], bnd[0])

        else:

            num, dist = graph.keys["right"], graph.dists["left"]
            y = numpy.sign(xloc)*numpy.abs(xloc)**(1./num)
            y[(xloc == 0.)*(num < 0)] = numpy.inf

            bnd = graph(y, dist)
            assert numpy.all((num % 1 == 0) + (bnd[0] >= 0)), \
                    "root of negative number"

            pair = num % 2 == 0
            bnd_ = numpy.empty(bnd.shape)
            bnd_[0] = numpy.where(pair*(bnd[0]*bnd[1]<0), 0, bnd[0])
            bnd_[0] = numpy.where(pair*(bnd[0]*bnd[1]>0), \
                    numpy.min(numpy.abs(bnd), 0), bnd_[0])**num
            bnd_[1] = numpy.where(pair, numpy.max(numpy.abs(bnd), 0),
                    bnd[1])**num

            bnd_[0], bnd_[1] = numpy.where(
                bnd_[0] < bnd_[1], bnd_[0], bnd_[1]
            ), numpy.where(
                bnd_[0] < bnd_[1], bnd_[1], bnd_[0]
            )

        return bnd_


    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        if "left" in graph.keys and "right" in graph.dists:

            num, dist = graph.keys["left"], graph.dists["right"]
            assert numpy.all(num>0), "imaginary result"

            y = numpy.log(numpy.abs(xloc) + 1.*(xloc<=0))/\
                    numpy.log(numpy.abs(num)+1.*(num == 1))

            out = graph(y, dist)
            out = numpy.where(xloc<=0, 0., out)

        else:

            num, dist = graph.keys["right"], graph.dists["left"]
            y = numpy.sign(xloc)*numpy.abs(xloc)**(1./num)
            pairs = numpy.sign(xloc**num) != -1

            _1 = graph.copy()(-y, dist)
            out = graph(y, dist)
            out = numpy.where(num < 0, 1-out, out - pairs*_1)

        return out

    def _ppf(self, q, graph):
        """Point percentile function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
            out = num**graph(q, dist)
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
            out = graph(q, dist)**num
        return out

    def _pdf(self, xloc, graph):
        """Probability density function."""
        if "left" in graph.keys and "right" in graph.dists:

            num, dist = graph.keys["left"], graph.dists["right"]
            assert numpy.all(num>0), "imaginary result"
            x_ = numpy.where(xloc<=0, -numpy.inf,
                    numpy.log(xloc + 1.*(xloc<=0))/numpy.log(num+1.*(num==1)))
            num_ = numpy.log(num+1.*(num==1))*xloc
            num_ = num_ + 1.*(num_==0)

            out = graph(x_, dist)/num_

        else:

            num, dist = graph.keys["right"], graph.dists["left"]
            x_ = numpy.sign(xloc)*numpy.abs(xloc)**(1./num -1)
            xloc = numpy.sign(xloc)*numpy.abs(xloc)**(1./num)
            pairs = numpy.sign(xloc**num) == 1

            G_ = graph.copy()
            out = graph(xloc, dist)
            if numpy.any(pairs):
                out = out + pairs*G_(-xloc, dist)
            out = numpy.sign(num)*out * x_ / num
            out[numpy.isnan(out)] = numpy.inf

        return out

    def _mom(self, k, graph):
        """Statistical moments."""
        if "right" in graph.keys and not numpy.any(graph.keys["right"] % 1):
            out = graph(k*numpy.array(graph.keys["right"], dtype=int), graph.dists["left"])
        else:
            raise NotImplementedError()
        return out

    def _val(self, graph):
        """Value extraction."""
        if len(graph.keys)==2:
            return graph.keys["left"]**graph.keys["right"]
        return self

    def _str(self, left, right):
        """String representation."""
        return "(%s)**(%s)" % (left,right)


def pow(left, right):
    """
    Power operator.

    Args:
        left (Dist, array_like) : Left hand side.
        right (Dist, array_like) : Right hand side.
    """
    if isinstance(left, Dist):

        if not isinstance(right, Dist):
            right = numpy.array(right)
            if right.size==1 and right==1:
                return left

    elif not isinstance(right, Dist):
        return left+right

    return Pow(left=left, right=right)



r"""
Constructing an transitional operator requires to construct the
distributions different from a stand-alone variable. To illustrate how
to construct these types of variables, consider the following example.
Let :math:`Q=A+B`, where one of :math:`A` and :math:`B` is a random
variable, and the other a scalar. Which variable is what dependents on
the user setup of the variable. Assuming that :math:`A` is the random
variable, we have that

.. math::

    F_{Q\mid B}(q\mid b) = \mathbb P {q\leq Q\mid B\!=\!b} =
    \mathbb P {q\leq AB\mid B\!=\!b}

    = \mathbb P {\tfrac qb\leq A\mid B\!=\!b} =
    F_{A\mid B}(\tfrac qb\mid b).

Because of symmetry the distribution will be the same, but with
:math:`A` and :math:`B` substituted.

This is required when trying to use operators on multivariate variables. To
create such a variable with ``construct`` provide an additional ``length``
keyword argument specifying the length of a distribution.
"""
from .baseclass import Graph



"""
Create all primes bellow a certain threshold.

Examples::

    >>> print(create_primes(1))
    []
    >>> print(create_primes(2))
    [2]
    >>> print(create_primes(3))
    [2, 3]
    >>> print(create_primes(20))
    [2, 3, 5, 7, 11, 13, 17, 19]
"""


def create_primes(threshold):
    """
    Generate prime values using sieve of Eratosthenes method.

    Args:
        threshold (int):
            The upper bound for the size of the prime values.

    Returns (List[int]):
        All primes from 2 and up to ``threshold``.
    """
    if threshold == 2:
        return [2]

    elif threshold < 2:
        return []

    numbers = list(range(3, threshold+1, 2))
    root_of_threshold = threshold ** 0.5
    half = int((threshold+1)/2-1)
    idx = 0
    counter = 3
    while counter <= root_of_threshold:
        if numbers[idx]:
            idy = int((counter*counter-3)/2)
            numbers[idy] = 0
            while idy < half:
                numbers[idy] = 0
                idy += counter
        idx += 1
        counter = 2*idx+3
    return [2] + [number for number in numbers if number]



"""
Generator for combining one dimensional rules into multivariate versions.
"""
import numpy

import chaospy.quad


def rule_generator(*funcs):
    """
    Constructor for creating multivariate quadrature generator.

    Args:
        *funcs (callable) : One dimensional integration rule where each rule
            returns `abscissas` and `weights` as one dimensional arrays. They
            must take one positional argument `order`.

    Returns:
        (callable) : Multidimensional integration quadrature function that
            takes the arguments `order` and `sparse`, and a optional `part`.
            The argument `sparse` is used to select for if Smolyak sparse grid
            is used, and `part` defines if subset of rule should be generated
            (for parallelization).

    Example:
        >>> clenshaw_curtis = lambda order: chaospy.quad_clenshaw_curtis(
        ...         order, lower=-1, upper=1, growth=True)
        >>> gauss_legendre = lambda order: chaospy.quad_gauss_legendre(
        ...         order, lower=0, upper=1)
        >>> quad_func = chaospy.rule_generator(clenshaw_curtis, gauss_legendre)
        >>> abscissas, weights = quad_func(1)
        >>> print(numpy.around(abscissas, 4))
        [[-1.     -1.      0.      0.      1.      1.    ]
         [ 0.2113  0.7887  0.2113  0.7887  0.2113  0.7887]]
        >>> print(numpy.around(weights, 4))
        [0.1667 0.1667 0.6667 0.6667 0.1667 0.1667]
    """
    dim = len(funcs)
    tensprod_rule = create_tensorprod_function(funcs)
    assert hasattr(tensprod_rule, "__call__")

    mv_rule = create_mv_rule(tensprod_rule, dim)
    assert hasattr(mv_rule, "__call__")
    return mv_rule


def create_tensorprod_function(funcs):
    """Combine 1-D rules into multivariate rule using tensor product."""
    dim = len(funcs)

    def tensprod_rule(order, part=None):
        """Tensor product rule."""
        order = order*numpy.ones(dim, int)
        values = [funcs[idx](order[idx]) for idx in range(dim)]

        abscissas = [numpy.array(_[0]).flatten() for _ in values]
        abscissas = chaospy.quad.combine(abscissas, part=part).T

        weights = [numpy.array(_[1]).flatten() for _ in values]
        weights = numpy.prod(chaospy.quad.combine(weights, part=part), -1)

        return abscissas, weights

    return tensprod_rule


def create_mv_rule(tensorprod_rule, dim):
    """Convert tensor product rule into a multivariate quadrature generator."""
    def mv_rule(order, sparse=False, part=None):
        """
        Multidimensional integration rule.

        Args:
            order (int, array_like) : order of integration rule. If array_like,
                order along each axis.
            sparse (bool) : use Smolyak sparse grid.

        Returns:
            (numpy.ndarray, numpy.ndarray) abscissas and weights.
        """
        if sparse:
            order = numpy.ones(dim, dtype=int)*order
            tensorprod_rule_ = lambda order, part=part:\
                tensorprod_rule(order, part=part)
            return chaospy.quad.sparse_grid(tensorprod_rule_, order)

        return tensorprod_rule(order, part=part)

    return mv_rule

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Use recursive method for calculating Fourier coefficients.
"""
import chaospy


class FourierRecursive(object):
    """Calculate Fourier coefficients using Bertran's recursive formula."""

    def __init__(self, dist):
        """
        Initiate module.

        Distribution to create orthogonal basis on
        coef = E[v[n]*P[i]*P[j]]/E[P[j]**2]
        where v is basis polynomial and P are orthogonal polynomials
        """
        self.dist = dist
        self.dim = len(dist)
        self.hist = {}

    def mom_111(self, idxi, idxj, idxk):
        """Backend moment for i, j, k == 1, 1, 1."""
        if (idxi, idxj, idxk) in self.hist:
            return self.hist[idxi, idxj, idxk]

        if idxi == idxk == 0 or idxi == idxj == 0:
            out = 0

        elif chaospy.bertran.add(idxi, idxk, self.dim) < idxj \
                or chaospy.bertran.add(idxi, idxj, self.dim) < idxk:
            out = 0

        elif chaospy.bertran.add(idxi, idxj, self.dim) == idxk:
            out = 1

        elif idxj == idxk == 0:
            out = self.dist.mom(chaospy.bertran.multi_index(idxi, self.dim))

        elif idxk == 0:
            out = self.mom_110(idxi, idxj, idxk)

        else:
            out = self.mom_recurse(idxi, idxj, idxk)

        self.hist[idxi, idxj, idxk] = out

        return out

    def mom_110(self, idxi, idxj, idxk):
        """Backend moment for i, j, k == 1, 1, 1."""
        rank_ = min(
            chaospy.bertran.rank(idxi, self.dim),
            chaospy.bertran.rank(idxj, self.dim),
            chaospy.bertran.rank(idxk, self.dim)
        )
        par, axis0 = chaospy.bertran.parent(idxj, self.dim)
        gpar, _ = chaospy.bertran.parent(par, self.dim, axis0)
        idxi_child = chaospy.bertran.child(idxi, self.dim, axis0)
        oneup = chaospy.bertran.child(0, self.dim, axis0)

        out = self(idxi_child, par, 0)
        for k in range(gpar, idxj):
            if chaospy.bertran.rank(k, self.dim) >= rank_:
                out -= self.mom_111(oneup, par, k) * self.mom_111(idxi, k, 0)
        return out

    def mom_recurse(self, idxi, idxj, idxk):
        """Backend mement main loop."""
        rank_ = min(
            chaospy.bertran.rank(idxi, self.dim),
            chaospy.bertran.rank(idxj, self.dim),
            chaospy.bertran.rank(idxk, self.dim)
        )
        par, axis0 = chaospy.bertran.parent(idxk, self.dim)
        gpar, _ = chaospy.bertran.parent(par, self.dim, axis0)
        idxi_child = chaospy.bertran.child(idxi, self.dim, axis0)
        oneup = chaospy.bertran.child(0, self.dim, axis0)

        out1 = self.mom_111(idxi_child, idxj, par)
        out2 = self.mom_111(
            chaospy.bertran.child(oneup, self.dim, axis0), par, par)
        for k in range(gpar, idxk):
            if chaospy.bertran.rank(k, self.dim) >= rank_:
                out1 -= self.mom_111(oneup, k, par) \
                    * self.mom_111(idxi, idxj, k)
                out2 -= self.mom_111(oneup, par, k) \
                    * self(oneup, k, par)
        return out1 / out2

    def __call__(self, idxi, idxj, idxk):
        """
        Calculate coefficient.

        idxi : int
            Single index for basis reference
        idxj : int
            Single index for for orthogonal poly in nominator
        idxk : int
            Single index for for orthogonal poly in denominator
        """
        return self.mom_111(idxi, idxj, idxk)


if __name__ == "__main__":
    import chaospy as cp
    import doctest
    doctest.testmod()



"""
Create samples from a Korobov lattice.

Examples usage
--------------

Normal usage::

    >>> print(create_korobov_samples(order=4, dim=2))
    [[0.2 0.4 0.6 0.8]
     [0.4 0.8 0.2 0.6]]

With custom number base::

    >>> print(create_korobov_samples(order=4, dim=2, base=3))
    [[0.2 0.4 0.6 0.8]
     [0.6 0.2 0.8 0.4]]
"""
import numpy


def create_korobov_samples(order, dim, base=17797):
    """
    Create Korobov lattice samples.

    Args:
        order (int):
            The order of the Korobov latice. Defines the number of
            samples.
        dim (int):
            The number of dimensions in the output.
        base (int):
            The number based used to calculate the distribution of values.

    Returns (numpy.ndarray):
        Korobov lattice with ``shape == (dim, order)``
    """
    values = numpy.empty(dim)
    values[0] = 1
    for idx in range(1, dim):
        values[idx] = base*values[idx-1] % (order+1)

    grid = numpy.mgrid[:dim, :order+1]
    out = values[grid[0]] * (grid[1]+1) / (order+1.) % 1.
    return out[:, :order]



"""
Hermite Genz-Keister quadrature rules

Adapted from John Burkardt's implementation in Matlab

Licensing
---------
This code is distributed under the GNU LGPL license.
"""

from .gk16 import quad_genz_keister_16
from .gk18 import quad_genz_keister_18
from .gk22 import quad_genz_keister_22
from .gk24 import quad_genz_keister_24

COLLECTION = {
    16: quad_genz_keister_16,
    18: quad_genz_keister_18,
    22: quad_genz_keister_22,
    24: quad_genz_keister_24,
}

from .genz_keister import quad_genz_keister



"""
Use of Bertran to calculate sparse segments.
"""
import numpy as np
import chaospy as cp


def sparse_segment(cords):
    r"""
    Create a segment of a sparse grid.

    More specifically, a segment of
    `\cup_{cords \in C} sparse_segment(cords) == sparse_grid(M)`
    where
    `C = {cords: M=sum(cords)}`

    Parameters
    ----------
    cords : array_like
        The segment to extract. `cord` must consist of non-negative intergers.

    Returns
    -------
    Q : ndarray
        Sparse segment where `Q.shape==(K, sum(M))` and `K` is segment
        specific.

    Convert a ol-index to sparse grid coordinates on [0, 1]^N hyper
    cube. A sparse grid of order `D` coencide with the set of
    sparse_segments where `||cords||_1 <= D`.

    Examples
    --------
    >>> print(cp.bertran.sparse_segment([0, 2]))
    [[0.5   0.125]
     [0.5   0.375]
     [0.5   0.625]
     [0.5   0.875]]

    >>> print(cp.bertran.sparse_segment([0, 1, 0, 0]))
    [[0.5  0.25 0.5  0.5 ]
     [0.5  0.75 0.5  0.5 ]]
    """
    cords = np.array(cords)+1
    slices = []
    for cord in cords:
        slices.append(slice(1, 2**cord+1, 2))

    grid = np.mgrid[slices]
    indices = grid.reshape(len(cords), np.prod(grid.shape[1:])).T
    sgrid = indices*2.**-cords
    return sgrid



# pylint: disable=protected-access
"""Backend for forward Rosenblatt transformations."""
import numpy


def fwd_call(self, x_data, dist):
    "forward call backend wrapper"
    self.counting(dist, "fwd")
    assert x_data.shape == (len(dist), self.size)
    self.dist, dist_ = dist, self.dist

    self.graph.add_node(dist, key=x_data)
    out = numpy.empty(x_data.shape)

    prm = self.dists.build()
    prm.update(self.keys.build())
    for key, value in prm.items():

        if not isinstance(value, numpy.ndarray):
            post_value = self.run(value, "val")[0]
            if isinstance(post_value, numpy.ndarray):
                prm[key] = post_value
                graph.add_node(value, key=post_value)

    if dist.advance:
        out[:] = dist._cdf(x_data, self)
    else:
        out[:] = dist._cdf(x_data, **prm)

    self.graph.add_node(dist, val=out)

    self.dist = dist_
    return numpy.array(out)



"""
Logarithm base N.
"""
import numpy

from ..baseclass import Dist


class Logn(Dist):
    """Logarithm base N."""

    def __init__(self, dist, n=2):
        """
        Constructor.

        Args:
            left (Dist, array_like) : Left hand side.
            right (Dist, array_like) : Right hand side.
        """
        assert isinstance(dist, Dist)
        assert numpy.all(dist.range()>=0)
        assert n>0 and n!=1
        Dist.__init__(self, dist=dist, n=n,
                _length=len(dist), _advance=True)

    def _str(self, dist, n):
        """String representation."""
        return "Logn(%s,&s)" % (dist, n)

    def _val(self, graph):
        """Value extraction."""
        if "dist" in graph.keys:
            return numpy.log(graph.keys["dist"])/numpy.log(graph.keys["n"])
        return self

    def _pdf(self, xloc, graph):
        """Probability density function."""
        n = graph.keys["n"]
        return graph(n**xloc, graph.dists["dist"])*numpy.log(n)*n**xloc

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        return graph(graph.keys["n"]**xloc, graph.dists["dist"])

    def _ppf(self, q, graph):
        """Point percentile function."""
        return numpy.log(graph(q, graph.dists["dist"]))/graph.keys["n"]

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        n = graph.keys["n"]
        lower,upper = graph(n**xloc, graph.dists["dist"])
        return numpy.log(lower)/numpy.log(n),\
                numpy.log(upper)/numpy.log(n)

def logn(dist, order):
    """
    Logarithm base N.

    Args:
        dist (Dist) : distribution (>=0)
        order (int) : logarithm base.
    """
    return Logn(dist, order)



import inspect

import numpy as np
from scipy.misc import comb

from . import collection, sparse_grid


def generate_quadrature(order, domain, accuracy=100, sparse=False, rule="C",
                        composite=1, growth=None, part=None, **kws):
    """
    Numerical quadrature node and weight generator.

    Args:
        order (int) : The order of the quadrature.
        domain (array_like, Dist) : If array is provided domain is the lower
        and upper bounds (lo,up). Invalid if gaussian is set.  If Dist is
            provided, bounds and nodes are adapted to the distribution. This
            includes weighting the nodes in Clenshaw-Curtis quadrature.
        accuracy (int) : If gaussian is set, but the Dist provieded in domain
            does not provide an analytical TTR, ac sets the approximation order
            for the descitized Stieltje's method.
        sparse (bool) : If True used Smolyak's sparse grid instead of normal
            tensor product grid.
        rule (str) : Rule for generating abscissas and weights. Either done
            with quadrature rules, or with random samples with constant
            weights.  For a description of the rules, see
            :py:`chaospy.quad.collection`.
        composite (int, optional) : If provided, composite quadrature will be
            used.  Value determines the number of domains along an axis.
            Ignored in the case gaussian=True.
        growth (bool, optional) : If True sets the growth rule for the
            composite quadrature rule to exponential for Clenshaw-Curtis
            quadrature.
        **kws (optional) : Extra keywords passed to ``generate_samples``.
    """
    if growth and order:
        if isinstance(order, int):
            order = 2**order
        else:
            order = tuple([2**o for o in order])

    from ..distributions.baseclass import Dist
    isdist = isinstance(domain, Dist)
    if isdist:
        dim = len(domain)
    else:
        dim = np.array(domain[0]).size

    rule = rule.lower()
    if len(rule) == 1:
        rule = collection.QUAD_SHORT_NAMES[rule]

    quad_function = collection.get_function(
        rule,
        domain,
        growth=growth,
        composite=composite,
        accuracy=accuracy,
    )

    if sparse:
        order = np.ones(len(domain), dtype=int)*order
        abscissas, weights = sparse_grid.sparse_grid(quad_function, order, dim)

    else:
        abscissas, weights = quad_function(order)

    assert len(weights) == abscissas.shape[1]
    assert len(abscissas.shape) == 2

    if isdist and not sparse:
        weights /= np.sum(weights)

    return abscissas, weights



"""
Custom distribution constructor.
"""
import types
from .baseclass import Dist

LEGAL_ATTRS = {
    "cdf": "_cdf", "bnd": "_bnd",
    "pdf": "_pdf", "ppf": "_ppf", "mom": "_mom",
    "ttr": "_ttr", "val": "_val", "doc": "__doc__",
    "str": "_str"
}

def construct(parent=None, defaults=None, advance=False, length=1, **kwargs):
    """
    Random variable constructor.

    Args:
        cdf (callable):
            Cumulative distribution function. Optional if ``parent`` is used.
        bnd (callable):
            Boundary interval. Optional if ``parent`` is used.
        parent (Dist):
            Distribution used as basis for new distribution. Any other argument
            that is omitted will instead take is function from ``parent``.
        doc (str, optional):
            Documentation for the distribution.
        str (str, callable, optional):
            Pretty print of the variable.
        pdf (callable, optional):
            Probability density function.
        ppf (callable, optional):
            Point percentile function.
        mom (callable, optional):
            Raw moment generator.
        ttr (callable, optional):
            Three terms recursion coefficient generator
        val (callable, optional):
            Value function for transferable distributions.
        dep (callable, optional):
            Dependency structure.
        advance (bool):
            If True, advance mode is used. See dist.graph for details.
        length (int):
            If constructing an multivariate random variable, this sets the
            assumed length. Defaults to 1.
        init (callable, optional):
            Custom constructor method.

    Returns:
        dist (Dist) : New custom distribution.
    """
    for key in kwargs:
        assert key in LEGAL_ATTRS, "{} is not legal input".format(key)

    if parent is not None:
        for key, value in LEGAL_ATTRS.items():
            if key not in kwargs and hasattr(parent, value):
                    kwargs[key] = getattr(parent, value)

    assert "cdf" in kwargs, "cdf function must be defined"
    assert "bnd" in kwargs, "bnd function must be defined"
    if "str" in kwargs and isinstance(kwargs["str"], str):
        string = kwargs.pop("str")
        kwargs["str"] = lambda *args, **kwargs: string

    defaults = defaults if defaults else {}
    for key in defaults:
        assert key in LEGAL_ATTRS, "invalid default value {}".format(key)
    defaults["_length"] = length
    defaults["_advance"] = advance

    def custom_distribution(**kws):

        prm = defaults.copy()
        prm.update(kws)
        dist = Dist(**prm)

        for key, function in kwargs.items():
            attr_name = LEGAL_ATTRS[key]
            setattr(dist, attr_name, types.MethodType(function, dist))
        return dist

    if "doc" in kwargs:
        custom_distribution.__doc__ = kwargs["doc"]

    return custom_distribution



# pylint: disable=protected-access
"""Backend for distribution dependency retriever."""
import numpy


def dep_call(self, dist):
    """Dependency call backend wrapper."""
    graph = self.graph
    self.dist, dist_ = dist, self.dist

    if hasattr(dist, "_dep"):
        out = dist._dep(self)
    else:
        for val in self.dist.prm.values():
            out = numpy.zeros(len(dist), dtype=bool)
            if val in graph.nodes:
                out = numpy.ones(len(dist), dtype=bool)
                break
            elif not isinstance(val, numpy.ndarray):
                graph.add_node(val)
    graph.add_node(dist, key=out)

    assert len(out) == len(dist)
    self.dist = dist_
    return out



# pylint: disable=protected-access
"""Backend for random number generator."""
import numpy


def rnd_call(self, dist):
    "Sample generator call backend wrapper"

    self.counting(dist, "rnd")
    graph = self.graph
    self.dist, dist_ = dist, self.dist

    for key, value in dist.prm.items():
        if not isinstance(value, numpy.ndarray) and\
                not "key" in graph.node[value]:
            self.rnd_call(value)

    if dist.advance:
        key, _ = self.run(dist, "val")
    else:
        rnd = numpy.random.random((len(dist), self.size))
        key = self.inv_call(rnd, dist)

    assert isinstance(key, numpy.ndarray)
    graph.add_node(dist, key=key)

    self.dist = dist_

    if dist is self.root:
        out = graph.node[dist]["key"]
        return out



"""
Multi-indexing tools after Bertran's notation

Polynomials can either be indexed with single index or multi-index
Bertran is a notation for switching between the two.
"""

import chaospy.bertran.operators
import chaospy.bertran.fourier
import chaospy.bertran.sparse

from chaospy.bertran.operators import *
from chaospy.bertran.sparse import sparse_segment



"""
Monte Carlo simulation is by nature a very slow converging method.  The error
in convergence is proportional to :math:`1/\sqrt{K}` where :math:`K` is the
number of samples.  It is somewhat better with variance reduction techniques
that often reaches errors proportional to :math:`1/K`. For a
full overview of the convergence rate of the various methods, see for
example :cite:`kroese_handbook_2011`. However as the number
of dimensions grows, Monte Carlo convergence rate stays the same, making it
immune to the curse of dimensionality.

Generating random samples can be done from the distribution instance
method ``sample`` as discussed in the :ref:`tutorial`.
For example, to generate nodes from the Korobov lattice::

    >>> distribution = chaospy.Iid(chaospy.Beta(2, 2), 2)
    >>> samples = distribution.sample(4, rule="K")
    >>> print(numpy.around(samples, 4))
    [[0.2871 0.4329 0.5671 0.7129]
     [0.4329 0.7129 0.2871 0.5671]]

At the core of `chaospy`, all samples are generated to the unit hyper-cube
through the function :func:`~chaospy.distributions.samplers.generator.generate_samples`.
These samples are then mapped using a :ref:`rosenblatt` to map the values on
the domain respective to the distribution in question. This way, all variance
reduction techniques are supported by all distributions.

If one wants even more control over the sample generation using low-discrepancy
sequencecs, it is also possible to take a look at
:mod:`chaospy.distributions.sampler.generator.sequences`. This module will contain all
the individual sequence schemes with extra flags for controlling each scheme.
In the case of the Krobov lattice::

    >>> from chaospy.distributions.sampler import create_korobov_samples
    >>> print(create_korobov_samples(order=4, dim=2))
    [[0.2 0.4 0.6 0.8]
     [0.4 0.8 0.2 0.6]]
    >>> print(create_korobov_samples(order=4, dim=2, base=1234))
    [[0.2 0.4 0.6 0.8]
     [0.8 0.6 0.4 0.2]]
"""
from .generator import generate_samples

from .sequences import *  # pylint: disable=wildcard-import
from .latin_hypercube import create_latin_hypercube_samples
from .antithetic import create_antithetic_variates



"""
Sparse grid constructor.

Method for turning collection of one dimensional quadrature rules into Smolyak
sparse grid tensor rules.
"""

import numpy
import scipy.misc

import chaospy.bertran


def sparse_grid(func, order, dim=None, skew=None):
    """
    Smolyak sparse grid constructor.

    Args:
        func (callable) : function that takes a single argument `order` of type
            `numpy` and with `order.shape = (dim,)`
        order (int, array_like) : The order of the grid. If `array_like`,
            it overrides both `dim` and `skew`.
        dim (int) : number of dimension.
        skew (list) : order skewness.
    """
    if not isinstance(order, int):
        orders = numpy.array(order).flatten()
        dim = orders.size
        m_order = int(numpy.min(orders))
        skew = [order-m_order for order in orders]
        return sparse_grid(func, m_order, dim, skew)

    abscissas, weights = [], []
    bindex = chaospy.bertran.bindex(order-dim+1, order, dim)

    if skew is None:
        skew = numpy.zeros(dim, dtype=int)
    else:
        skew = numpy.array(skew, dtype=int)
        assert len(skew) == dim

    for idx in range(
            chaospy.bertran.terms(order, dim)
            - chaospy.bertran.terms(order-dim, dim)):

        idb = bindex[idx]
        abscissa, weight = func(skew+idb)
        weight *= (-1)**(order-sum(idb))*scipy.misc.comb(dim-1, order-sum(idb))
        abscissas.append(abscissa)
        weights.append(weight)

    abscissas = numpy.concatenate(abscissas, 1)
    weights = numpy.concatenate(weights, 0)

    abscissas = numpy.around(abscissas, 15)
    order = numpy.lexsort(tuple(abscissas))
    abscissas = abscissas.T[order].T
    weights = weights[order]

    # identify non-unique terms
    diff = numpy.diff(abscissas.T, axis=0)
    unique = numpy.ones(len(abscissas.T), bool)
    unique[1:] = (diff != 0).any(axis=1)

    # merge duplicate nodes
    length = len(weights)
    idx = 1
    while idx < length:
        while idx < length and unique[idx]:
            idx += 1
        idy = idx+1
        while idy < length and not unique[idy]:
            idy += 1
        if idy-idx > 1:
            weights[idx-1] = numpy.sum(weights[idx-1:idy])
        idx = idy+1

    abscissas = abscissas[:, unique]
    weights = weights[unique]

    return abscissas, weights



"""
"""

import numpy as np
from . import approx


def fwd(dist, x):
    """Forward Rosenblatt transformation."""
    dim = len(dist)
    x = np.asfarray(x)
    shape = x.shape
    size = int(x.size/dim)
    x = x.reshape(dim, size)

    bnd, graph = dist.graph.run(x, "range")
    x_ = np.where(x < bnd[0], bnd[0], x)
    x_ = np.where(x_ > bnd[1], bnd[1], x_)
    out, graph = dist.graph.run(x_, "fwd")
    out = np.where(x < bnd[0], 0, out)
    out = np.where(x > bnd[1], 1, out)

    out = out.reshape(shape)
    return out


def inv(dist, q, maxiter=100, tol=1e-5, **kws):
    """Inverse Rosenblatt transformation."""
    q = np.array(q)
    assert np.all(q>=0) and np.all(q<=1), q

    dim = len(dist)
    shape = q.shape
    size = int(q.size/dim)
    q = q.reshape(dim, size)

    try:
        out, graph = dist.graph.run(q, "inv", maxiter=maxiter, tol=tol)

    except NotImplementedError:
        out, N, q_ = approx.inv(dist, q, maxiter=maxiter, tol=tol, retall=True)

        diff = np.max(np.abs(q-q_))
        print("approx %s.inv w/%d calls and eps=%g" % (dist, N, diff))

    lo,up = dist.graph.run(out, "range")[0]
    out = np.where(out.T>up.T, up.T, out.T).T
    out = np.where(out.T<lo.T, lo.T, out.T).T
    out = out.reshape(shape)

    return out



"""
Descriptives are a collection of statistical analysis tools that can be used to
analyse :ref:`distributions` and :ref:`polynomials`, both as an expansion (see
:ref:`orthogonality`) and as results (see :ref:`regression` and
:ref:`spectral`).  For example, the expected value operator
:func:`~chaospy.descriptives.E` can be applied on distributions directly as
follows::

    >>> distribution = cp.Uniform(0, 1)
    >>> expected = cp.E(distribution)
    >>> print(expected)
    0.5

For multivariate distributions::

    >>> distribution = cp.J(
    ...     cp.Uniform(0, 1),
    ...     cp.Normal(0, 1)
    ... )
    >>> expected = cp.E(distribution)
    >>> print(expected)
    [0.5 0. ]


For simple polynomials, distribution goes as the second argument. In other
words, it calculates the expected value of the unit variable with respect to
distribution. For example::

    >>> distribution = cp.J(
    ...     cp.Uniform(0, 1),
    ...     cp.Normal(0, 1)
    ... )
    >>> q0, q1 = cp.variable(2)
    >>> expected = cp.E(q1, distribution)
    >>> print(expected)
    0.0
"""
import chaospy as cp

import chaospy.descriptives.first
import chaospy.descriptives.second1d
import chaospy.descriptives.second2d
import chaospy.descriptives.higher
import chaospy.descriptives.sensitivity
import chaospy.descriptives.misc

from chaospy.descriptives.first import E, E_cond
from chaospy.descriptives.second1d import Var, Std
from chaospy.descriptives.second2d import Cov, Corr
from chaospy.descriptives.higher import Skew, Kurt

from chaospy.descriptives.sensitivity import (
    Sens_m, Sens_m2, Sens_m_nataf, Sens_nataf, Sens_t, Sens_t_nataf,
)
from chaospy.descriptives.misc import Acf, Spearman, Perc, QoI_Dist



"""
Collection of mathematical operators.
"""
from .addition import add
from .logarithm import log
from .logarithm10 import log10
from .logarithmn import logn
from .multiply import mul
from .negative import neg
from .power import pow
from .trunkation import trunk

from .trignometric import (
    Sin, Arcsin, Cos, Arccos, Tan, Arctan,
)
from .hyperbolic import (
    Sinh, Arcsinh, Cosh, Arccosh, Tanh, Arctanh
)



"""
Collection of approximation methods

Global methods are used on a distribution as a wrapper. Local
function are used by the graph-module as part of calculations.

Functions
---------
    pdf         Probability density function (local)
    pdf_full    Probability density function (global)
    ppf         Inverse CDF (local)
    inv         Inverse CDF (global)
    mom         Raw statistical moments (global)
    find_interior_point     Find an interior point (global)
"""
import numpy

import chaospy.quad

from .baseclass import Dist


def pdf(dist, x, G, eps=1.e-7, verbose=False,
        retall=False):
    """
Calculate the probability density function locally.

Parameters
----------
dist : Dist
    Distribution in question. May not be an advanced variable.
x : numpy.ndarray
    Location coordinates. Requires that x.shape=(len(dist), K).
G : Graph
    The chaospy state of the distribution calculations.
eps : float
    Acceptable error level for the approximations
retall : bool
    If True return Graph with the next calculation state with the
    approximation.

Returns
-------
out[, G]

out : numpy.ndarray
    Local probability density function with out.shape=x.shape.
    To calculate actual density function: numpy.prod(out, 0)
G : Graph
    The chaospy calculation state after approximation is complete.
    """

    x = numpy.asfarray(x)
    lo,up = numpy.min(x), numpy.max(x)
    mu = .5*(lo+up)
    eps = numpy.where(x<mu, eps, -eps)

    G.__call__ = G.fwd_call

    out = numpy.empty(x.shape)
    for d in range(len(dist)):
        x[d] += eps[d]
        out[d] = G.copy()(x.copy(), dist)[d]
        x[d] -= eps[d]

    out = numpy.abs((out-G(x.copy(), dist))/eps)

    G.__call__ = G.pdf_call

    if retall:
        return out, G
    return out

def pdf_full(dist, x, eps=1.e-7, verbose=False,
        retall=False):
    """
Calculate the probability density function globaly.

Parameters
----------
dist : Dist
    Distribution in question. May not be an advanced variable.
x : numpy.ndarray
    Location coordinates. Requires that x.shape=(len(dist), K).
eps : float
    Acceptable error level for the approximations
retall : bool
    If True return Graph with the next calculation state with the
    approximation.

Returns
-------
out[, G]

out : numpy.ndarray
    Global probability density function with out.shape=x.shape.
    To calculate actual density function: numpy.prod(out, 0)
G : Graph
    The chaospy calculation state after approximation is complete.
    """

    dim = len(dist)
    x = numpy.asfarray(x)

    shape = x.shape
    x = x.reshape(dim, x.size/dim)
    xdx = x.copy()
    y,G = dist.fwd(x,retall=True)
    lo,up = dist.range(x)
    mu = .5*(lo+up)
    out = numpy.empty(shape)
    eps = eps*numpy.ones(dim)

    for i in range(dim):
        eps_ = numpy.where(x[i]<mu[i], eps[i], -eps[i])

        xdx[i] += eps_
        out[i] = numpy.abs(dist.fwd(xdx)[i]-y[i])/eps[i]
        xdx[i] -= eps_

    if retall:
        return out, G
    return out

def inv(dist, q, maxiter=100, tol=1e-5, retall=False,
        verbose=False):
    """
Calculate the approximation of the point percentile function.

Parameters
----------
dist : Dist
    Distribution to estimate ppf.
q : numpy.ndarray
    Input values. All values must be on [0,1] and
    `q.shape==(dim,size)` where dim is the number of dimensions in
    dist and size is the number of values to calculate
    simultaneously.
maxiter : int
    The maximum number of iterations allowed before aborting
tol : float
    Tolerance parameter determining convergence.
retall : bool
    If true, return all.

Returns
-------
x, itrs, y

x : numpy.ndarray
    Distribution definition values.
itrs : int
    The number of iterations used before converging.
y : numpy.ndarray
    The model forward transformed value in x
    """

    dim = len(dist)
    size = q.size/dim
    q = q.reshape(dim, size)
    lo,up = dist.range(numpy.zeros((dim, size)))
    lo = lo*numpy.ones((dim,size))
    up = up*numpy.ones((dim,size))

    span = .5*(up-lo)
    too_much = numpy.any(dist.fwd(lo)>0, 0)
    while numpy.any(too_much):
        lo[:,too_much] -= span[:,too_much]
        too_much[too_much] = numpy.any(dist.fwd(lo)[:,too_much]>0, 0)

    too_little = numpy.any(dist.fwd(up)<1, 0)
    while numpy.any(too_little):
        up[:, too_little] += span[:, too_little]
        too_little[too_little] = numpy.any(dist.fwd(up)[:,too_little]<1, 0)

    # Initial values
    x = (up-lo)*q + lo
    flo, fup = -q, 1-q
    fx = tol*10*numpy.ones((dim,size))
    div = numpy.any((x<up)*(x>lo), 0)

    for iteration in range(1, maxiter+1):

        # eval function
        fx[:,div] = dist.fwd(x)[:,div]-q[:,div]

        # convergence test
        div[div] = numpy.any(numpy.abs(fx)>tol, 0)[div]
        if not numpy.any(div):
            break

        dfx = dist.pdf(x)[:,div]
        dfx = numpy.where(dfx==0, numpy.inf, dfx)

        # reduce boundaries
        lo_,up_ = dist.range(x)
        flo[:,div] = numpy.where(fx<=0, fx, flo)[:,div]
        lo[:,div] = numpy.where(fx<=0, x, lo)[:,div]
        lo = numpy.min([lo_, lo], 0)

        fup[:,div] = numpy.where(fx>=0, fx, fup)[:,div]
        up[:,div] = numpy.where(fx>=0, x, up)[:,div]
        up = numpy.max([up_, up], 0)

        # Newton increment
        xdx = x[:,div]-fx[:,div]/dfx

        # if new val on interior use Newton
        # else binary search
        x[:,div] = numpy.where((xdx<up[:,div])*(xdx>lo[:,div]),
                xdx, .5*(up+lo)[:,div])


    if retall:
        return x, iteration, dist.fwd(x)
    return x


def ppf(dist, q, G, maxiter=100, tol=1e-5, retall=False,
        verbose=False):
    """
Calculate the approximation of the point percentile function.

Parameters
----------
dist : Dist
    Distribution to estimate ppf.
q : numpy.ndarray
    Input values. All values must be on [0,1] and
    `q.shape==(dim,size)` where dim is the number of dimensions in
    dist and size is the number of values to calculate
    simultaneously.
maxiter : int
    The maximum number of iterations allowed before aborting
tol : float
    Tolerance parameter determining convergence.
retall : bool
    If true, return all.

Returns
-------
x, itrs

x : numpy.ndarray
    Distribution definition values.
itrs : int
    The number of iterations used before converging.
    """

    if not dist.advance:
        dist.prm, prm = G.K.build(), dist.prm
        out = inv(dist, q, maxiter, tol, retall, verbose)
        dist.prm = prm
        return out

    dim = len(dist)
    shape = q.shape
    size = q.size/dim
    q = q.reshape(dim, size)

    X = G.copy().run(size, "rnd")[1].node[dist]["key"]

    x = numpy.mean(X, -1)
    lo,up = numpy.min(X, -1), numpy.max(X, -1)
    lo = (lo*numpy.ones((size,dim))).T
    up = (up*numpy.ones((size,dim))).T

    # Initial values
    x = ((up.T-lo.T)*q.T + lo.T).T
    flo, fup = -q, 1-q
    fx = Fx = tol*10*numpy.ones((dim,size))
    dfx = 1.

    for iteration in range(1, maxiter+1):

        try:
            # eval function
            fx = G.copy().fwd_call(x, dist)
            success = (fx>=0)*(fx<=1)
            Fx = fx-q

            dfx = G.copy().pdf_call(x, dist)
            dfx = numpy.where(dfx==0, numpy.inf, dfx)

        except:
            success = numpy.zeros(size, dtype=bool)

        # convergence test
        if numpy.all(success) and numpy.all(numpy.abs(fx)<tol):
            break

        # reduce boundaries
        flo = numpy.where((Fx<0)*success, Fx, flo)
        lo = numpy.where((Fx<0)*success, x, lo)

        fup = numpy.where((Fx>0)*success, Fx, fup)
        up = numpy.where((Fx>0)*success, x, up)

        # Newton increment
        xdx = x-Fx/dfx

        # if new val on interior use Newton
        # else binary search
        x = numpy.where(success, xdx, .5*(up+lo))

    x = x.reshape(shape)

    if retall:
        return x, iteration, Fx
    return x


def mom(dist, K, retall=False, control_var=None,
        **kws):
    """
Approxmethod for estimation of raw statistical moments.

Parameters
----------
dist : Dist
    Distribution domain with dim=len(dist)
K : numpy.ndarray
    The exponents of the moments of interest with shape (dim,K).

Optional keywords

control_var : Dist
    If provided will be used as a control variable to try to reduce
    the error.
acc : int, optional
    The order of quadrature/MCI
sparse : bool
    If True used Smolyak's sparse grid instead of normal tensor
    product grid in numerical integration.
rule : str
    Quadrature rule
    Key     Description
    ----    -----------
    "G"     Optiomal Gaussian quadrature from Golub-Welsch
            Slow for high order and composit is ignored.
    "E"     Gauss-Legendre quadrature
    "C"     Clenshaw-Curtis quadrature. Exponential growth rule is
            used when sparse is True to make the rule nested.

    Monte Carlo Integration
    Key     Description
    ----    -----------
    "H"     Halton sequence
    "K"     Korobov set
    "L"     Latin hypercube sampling
    "M"     Hammersley sequence
    "R"     (Pseudo-)Random sampling
    "S"     Sobol sequence

composit : int, array_like optional
    If provided, composit quadrature will be used.
    Ignored in the case if gaussian=True.

    If int provided, determines number of even domain splits
    If array of ints, determines number of even domain splits along
        each axis
    If array of arrays/floats, determines location of splits

antithetic : array_like, optional
    List of bool. Represents the axes to mirror using antithetic
    variable during MCI.
    """

    dim = len(dist)
    shape = K.shape
    size = int(K.size/dim)
    K = K.reshape(dim,size)

    if dim>1:
        shape = shape[1:]

    order = kws.pop("order", 40)
    X,W = chaospy.quad.generate_quadrature(order, dist, **kws)


    grid = numpy.mgrid[:len(X[0]),:size]
    X = X.T[grid[0]].T
    K = K.T[grid[1]].T
    out = numpy.prod(X**K, 0)*W

    if not (control_var is None):

        Y = control_var.ppf(dist.fwd(X))
        mu = control_var.mom(numpy.eye(len(control_var)))

        if mu.size==1 and dim>1:
            mu = mu.repeat(dim)

        for d in range(dim):
            alpha = numpy.cov(out, Y[d])[0,1]/numpy.var(Y[d])
            out -= alpha*(Y[d]-mu)

    out = numpy.sum(out, -1)

    return out


def find_interior_point(dist):
    """
Find interior point using the range-function

Parameters
----------
dist : Dist
    Distribution to find interior on.

Returns
-------
interior_point : numpy.ndarray
    shape=(len(dist),)
    """
    try:
        x = dist.inv([.5]*len(dist))
        return x
    except:
        pass

    bnd = dist.range(numpy.zeros(len(dist)))
    x = .5*(bnd[1]-bnd[0])

    for i in range(10):
        bnd = dist.range(x)
        x_ = .5*(bnd[1]-bnd[0])
        if numpy.allclose(x, x_):
            break
        x = x_

    return x

# TODO: integrate these two functions.
def ttr(order, domain, **kws):

    prm = kws
    prm["accuracy"] = order
    prm["retall"] = True

    def _three_terms_recursion(self, keys, **kws):
        _, _, coeffs1, coeffs2 = chaospy.quad.generate_stieltjes(
            domain, numpy.max(keys)+1, **self1.prm)
        out = numpy.ones((2,) + keys.shape)
        idx = 0
        for idzs in keys.T:
            idy = 0
            for idz in idzs:
                if idz:
                    out[:, idy, idx] = coeffs1[idy, idz], coeffs2[idy, idz]
                idy += 1
            idx += 1

    return _three_terms_recursion

def moment_generator(order, domain, accuracy=100, sparse=False, rule="C",
                     composite=1, part=None, trans=lambda x:x, **kws):
    """Moment generator."""
    if isinstance(domain, Dist):
        dim = len(domain)
    else:
        dim = numpy.array(domain[0]).size

    if not numpy.array(trans(numpy.zeros(dim))).shape:
        func = trans
        trans = lambda x: [func(x)]

    if part is None:

        abscissas, weights = chaospy.quad.generate_quadrature(
            order, domain=domain, accuracy=accuracy, sparse=sparse,
            rule=rule, composite=composite, part=part, **kws)
        values = numpy.transpose(trans(abscissas))

        def moment_function(keys):
            """Raw statistical moment function."""
            return numpy.sum(numpy.prod(values**keys, -1)*weights, 0)
    else:

        isdist = isinstance(domain, Dist)
        if isdist:
            lower, upper = domain.range()
        else:
            lower, upper = numpy.array(domain)

        abscissas = []
        weights = []
        values = []
        for idx in numpy.ndindex(*part):
            abscissa, weight = chaospy.quad.collection.clenshaw_curtis(
                order, lower, upper, part=(idx, part))
            value = numpy.array(trans(abscissa))

            if isdist:
                weight *= domain.pdf(abscissa).flatten()

            if numpy.any(weight):
                abscissas.append(abscissa)
                weights.append(weight)
                values.append(value)

        def moment_function(keys):
            """Raw statistical moment function."""
            out = 0.
            for idx in range(len(abscissas)):
                out += numpy.sum(
                    numpy.prod(values[idx].T**keys, -1)*weights[idx], 0)
            return out

    def mom(keys, **kws):
        """Statistical moment function."""
        return numpy.array([moment_function(key) for key in keys.T])

    return mom



"""
Preprocessing of Poly class arguments.
"""
import numpy as np

import chaospy.poly.constructor


def preprocess(core, dim, shape, dtype):
    """Constructor function for the Poly class."""
    core, dim_, shape_, dtype_ = chaospy.poly.constructor.identify_core(core)

    core, shape = chaospy.poly.constructor.ensure_shape(core, shape, shape_)
    core, dtype = chaospy.poly.constructor.ensure_dtype(core, dtype, dtype_)
    core, dim = chaospy.poly.constructor.ensure_dim(core, dim, dim_)

    # Remove empty elements
    for key in list(core.keys()):
        if np.all(core[key] == 0):
            del core[key]

    assert isinstance(dim, int), \
        "not recognised type for dim: '%s'" % repr(type(dim))
    assert isinstance(shape, tuple), str(shape)
    assert dtype is not None, str(dtype)

    # assert non-empty container
    if not core:
        core = {(0,)*dim: np.zeros(shape, dtype=dtype)}
    else:
        core = {key: np.asarray(value, dtype) for key, value in core.items()}

    return core, dim, shape, dtype
